""" Kerr Nonlinear Oscillator """

from flax import struct
from jax import config

import jax.numpy as jnp

from jaxquantum.devices.base.base import Device, BasisTypes, HamiltonianTypes
from jaxquantum.core.operators import identity, destroy, create

config.update("jax_enable_x64", True)


@struct.dataclass
class KNO(Device):
    """
    Kerr Nonlinear Oscillator Device.
    """

    @classmethod
    def param_validation(cls, N, N_pre_diag, params, hamiltonian, basis):
        """ This can be overridden by subclasses."""
        assert basis == BasisTypes.fock, "Kerr Nonlinear Oscillator must be defined in the Fock basis." 
        assert hamiltonian == HamiltonianTypes.full, "Kerr Nonlinear Oscillator uses a full Hamiltonian."
        assert "ω" in params and "α" in params, "Kerr Nonlinear Oscillator requires frequency 'ω' and anharmonicity 'α' as parameters."
        
    def common_ops(self):
        ops = {}

        N = self.N
        ops["id"] = identity(N)
        ops["a"] = destroy(N)
        ops["a_dag"] = create(N)
        ops["phi"] = (ops["a"] + ops["a_dag"]) / jnp.sqrt(2)
        ops["n"] = 1j * (ops["a_dag"] - ops["a"]) / jnp.sqrt(2)
        return ops

    def get_linear_ω(self):
        """Get frequency of linear terms."""
        return self.params["ω"]

    def get_anharm(self):
        """Get anharmonicity."""
        return self.params["α"]

    def get_H_linear(self):
        """Return linear terms in H."""
        w = self.get_linear_ω()
        return w * self.linear_ops["a_dag"] @ self.linear_ops["a"]

    def get_H_full(self):
        """Return full H in linear basis."""
        α = self.get_anharm()

        return self.get_H_linear() + (α / 2) * (
            self.linear_ops["a_dag"]
            @ self.linear_ops["a_dag"]
            @ self.linear_ops["a"]
            @ self.linear_ops["a"]
        )




""" Flux base device."""

from abc import abstractmethod, ABC
from enum import Enum
from typing import Dict, Any, List

from flax import struct
from jax import config, Array
import jax.numpy as jnp
import matplotlib.pyplot as plt

from jaxquantum.devices.common.utils import harm_osc_wavefunction
from jaxquantum.devices.base.base import Device, BasisTypes

config.update("jax_enable_x64", True)

@struct.dataclass
class FluxDevice(Device):
    @abstractmethod
    def phi_zpf(self):
        """Return Phase ZPF."""

    def _calculate_wavefunctions_fock(self, phi_vals):
        """Calculate wavefunctions at phi_exts."""
        phi_osc = self.phi_zpf() * jnp.sqrt(2) # length of oscillator
        phi_vals = jnp.array(phi_vals)

        # calculate basis functions
        basis_functions = []
        for n in range(self.N_pre_diag):
            basis_functions.append(
                harm_osc_wavefunction(n, phi_vals, jnp.real(phi_osc))
            )
        basis_functions = jnp.array(basis_functions)

        # transform to better diagonal basis
        basis_functions_in_H_eigenbasis = self.get_vec_data_in_H_eigenbasis(basis_functions)
        
        # the below is equivalent to evecs_in_H_eigenbasis @ basis_functions_in_H_eigenbasis
        # since evecs in H_eigenbasis is diagonal, i.e. the identity matrix
        wavefunctions = basis_functions_in_H_eigenbasis 
        return wavefunctions
    
    def _calculate_wavefunctions_charge(self, phi_vals):
        phi_vals = jnp.array(phi_vals)

        # calculate basis functions
        basis_functions = []
        n_max = (self.N_pre_diag - 1) // 2
        for n in jnp.arange(-n_max, n_max + 1):
            basis_functions.append(
                1/(jnp.sqrt(2*jnp.pi)) * jnp.exp(1j * n * (2*jnp.pi*phi_vals))
            )
        basis_functions = jnp.array(basis_functions)

        # transform to better diagonal basis
        basis_functions_in_H_eigenbasis = self.get_vec_data_in_H_eigenbasis(basis_functions)
        
        # the below is equivalent to evecs_in_H_eigenbasis @ basis_functions_in_H_eigenbasis
        # since evecs in H_eigenbasis is diagonal, i.e. the identity matrix
        phase_correction_factors =  (1j**(jnp.arange(0,self.N_pre_diag))).reshape(self.N_pre_diag,1) # TODO: review why these are needed...
        wavefunctions = basis_functions_in_H_eigenbasis * phase_correction_factors
        return wavefunctions
    
    @abstractmethod
    def potential(self, phi):
        """Return potential energy as a function of phi."""

    def plot_wavefunctions(self, phi_vals, max_n=None, which=None, ax=None, mode="abs"):

        if self.basis == BasisTypes.fock:
            _calculate_wavefunctions = self._calculate_wavefunctions_fock
        elif self.basis == BasisTypes.charge:
            _calculate_wavefunctions = self._calculate_wavefunctions_charge
        else:
            raise NotImplementedError(f"The {self.basis} is not yet supported for plotting wavefunctions.")

        """Plot wavefunctions at phi_exts."""
        wavefunctions = _calculate_wavefunctions(phi_vals)
        energy_levels = self.eig_systems["vals"][:self.N]

        potential = self.potential(phi_vals)

        if ax is None:
            fig, ax = plt.subplots(1,1, figsize = (3.5, 2.5), dpi = 1000)
        else:
            fig = ax.get_figure()

        min_val = None
        max_val = None

        assert max_n is None or which is None, "Can't specify both max_n and which"

        max_n = self.N if max_n is None else max_n
        levels = range(max_n) if which is None else which

        for n in levels:
            if mode == "abs":
                wf_vals = jnp.abs(wavefunctions[n, :])**2
            elif mode == "real":
                wf_vals = wavefunctions[n, :].real
            elif mode == "imag":
                wf_vals = wavefunctions[n, :].imag

            wf_vals += energy_levels[n]
            curr_min_val = min(wf_vals)
            curr_max_val = max(wf_vals)

            if min_val is None or curr_min_val < min_val:
                min_val = curr_min_val

            if max_val is None or curr_max_val > max_val:
                max_val = curr_max_val

            ax.plot(phi_vals, wf_vals, label=f"$|${n}$\\rangle$", linestyle = '-', linewidth = 1)
            ax.fill_between(phi_vals, energy_levels[n], wf_vals, alpha=0.5)
        
        ax.plot(phi_vals, potential, label="potential", color="black", linestyle = '-', linewidth = 1)

        ax.set_ylim([min_val-1, max_val+1])

        ax.set_xlabel(r"$\Phi/\Phi_0$")
        ax.set_ylabel(r"Energy [GHz]")

        if mode == "abs":
            title_str = r"$|\psi_n(\Phi)|^2$"
        elif mode == "real":
            title_str = r"Re($\psi_n(\Phi)$)"
        elif mode == "imag":
            title_str = r"Im($\psi_n(\Phi)$)"

        ax.set_title(f"{title_str}")

        plt.legend(fontsize=6)
        fig.tight_layout()

        return ax




"""
jaxquantum
"""

import os

from .utils import *
from .core import *


with open(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "VERSION.txt")), "r"
) as _ver_file:
    __version__ = _ver_file.read().rstrip()

__author__ = "Shantanu Jha"
__credits__ = "EQuS"



""" Devices. """

from .flux_base import *
from .drive import *

from .resonator import *
from .kno import *
from .transmon import *
from .tunable_transmon import *
from .fluxonium import *
from .ats import *
from .ideal_qubit import *



""" IdealQubit."""

from flax import struct
from jax import config


from jaxquantum.devices.base.base import Device, BasisTypes, HamiltonianTypes
from jaxquantum.core.operators import identity, sigmaz, sigmax, sigmay, sigmam, sigmap

config.update("jax_enable_x64", True)


@struct.dataclass
class IdealQubit(Device):
    """
    Ideal qubit Device.
    """

    @classmethod
    def param_validation(cls, N, N_pre_diag, params, hamiltonian, basis):
        """ This can be overridden by subclasses."""
        assert basis == BasisTypes.fock, "IdealQubit is a two-level system defined in the Fock basis." 
        assert hamiltonian == HamiltonianTypes.full, "IdealQubit requires a full Hamiltonian."
        assert N == N_pre_diag == 2, "IdealQubit is a two-level system."
        assert "ω" in params, "IdealQubit requires a frequency parameter 'ω'."

    def common_ops(self):
        """Written in the linear basis."""
        ops = {}

        assert self.N_pre_diag == 2
        assert self.N == 2

        N = self.N_pre_diag
        ops["id"] = identity(N)
        ops["sigmaz"] = sigmaz()
        ops["sigmax"] = sigmax()
        ops["sigmay"] = sigmay()
        ops["sigmam"] = sigmam()
        ops["sigmap"] = sigmap()

        return ops

    def get_linear_ω(self):
        """Get frequency of linear terms."""
        return self.params["ω"]

    def get_H_linear(self):
        """Return linear terms in H."""
        w = self.get_linear_ω()
        return (w / 2) * self.linear_ops["sigma_z"]

    def get_H_full(self):
        """Return full H in linear basis."""
        return self.get_H_linear()


"""
Base Bosonic Qubit Class
"""

from typing import Dict, Optional, Tuple
from abc import abstractmethod, ABCMeta

from jaxquantum.utils.utils import device_put_params
import jaxquantum as jqt

from jax import config
import numpy as np
import jax.numpy as jnp
import jax.scipy as jsp
import matplotlib.pyplot as plt

config.update("jax_enable_x64", True)


class BosonicQubit(metaclass=ABCMeta):
    """
    Base class for Bosonic Qubits.
    """
    name = "bqubit"

    @property
    def _non_device_params(self):
        """
        Can be overriden in child classes.
        """
        return ["N"]

    def __init__(self, params: Optional[Dict[str, float]] = None, name: str = None):
        if name is not None:
            self.name = name

        self.params = params if params else {}
        self._params_validation()

        self.params = device_put_params(self.params, self._non_device_params)

        self.common_gates: Dict[str, jqt.Qarray] = {}
        self._gen_common_gates()

        self.wigner_pts = jnp.linspace(-4.5, 4.5, 61)

        self.basis = self._get_basis_states()

        for basis_state in ["+x", "-x", "+y", "-y", "+z", "-z"]:
            assert (
                basis_state in self.basis
            ), f"Please set the {basis_state} basis state."

    def _params_validation(self):
        """
        Override this method to add additional validation to params.

        E.g.
        if "N" not in self.params:
            self.params["N"] = 50
        """
        if "N" not in self.params:
            self.params["N"] = 50

    def _gen_common_gates(self):
        """
        Override this method to add additional common gates.

        E.g.
        if "N" not in self.params:
            self.params["N"] = 50
        """
        N = self.params["N"]
        self.common_gates["a_dag"] = jqt.create(N)
        self.common_gates["a"] = jqt.destroy(N)

    @abstractmethod
    def _get_basis_z(self) -> Tuple[jqt.Qarray, jqt.Qarray]:
        """
        Returns:
            plus_z (jqt.Qarray), minus_z (jqt.Qarray): z basis states
        """

    def _get_basis_states(self) -> Dict[str, jqt.Qarray]:
        """
        Construct basis states |+-x>, |+-y>, |+-z>
        """
        plus_z, minus_z = self._get_basis_z()
        return self._gen_basis_states_from_z(plus_z, minus_z)

    def _gen_basis_states_from_z(
        self, plus_z: jqt.Qarray, minus_z: jqt.Qarray
    ) -> Dict[str, jqt.Qarray]:
        """
        Construct basis states |+-x>, |+-y>, |+-z> from |+-z>
        """
        basis: Dict[str, jqt.Qarray] = {}
        N = self.params["N"]

        # import to make sure that each basis state is a column vec
        # otherwise, transposing a 1D vector will do nothing

        basis["+z"] = plus_z
        basis["-z"] = minus_z

        basis["+x"] = jqt.unit(basis["+z"] + basis["-z"])
        basis["-x"] = jqt.unit(basis["+z"] - basis["-z"])
        basis["+y"] = jqt.unit(basis["+z"] + 1j * basis["-z"])
        basis["-y"] = jqt.unit(basis["+z"] - 1j * basis["-z"])
        return basis

    def jqt2qt(self, state):
        return jqt.jqt2qt(state)

    # gates
    # ======================================================
    # @abstractmethod
    # def stabilize(self) -> None:
    #     """
    #     Stabilizing/measuring syndromes.
    #     """

    @property
    def x_U(self) -> jqt.Qarray:
        """
        Logical X unitary gate.
        """
        return self._gen_pauli_U("x")

    @property
    def x_H(self) -> Optional[jqt.Qarray]:
        """
        Logical X hamiltonian.
        """
        return None

    @property
    def y_U(self) -> jqt.Qarray:
        """
        Logical Y unitary gate.
        """
        return self._gen_pauli_U("y")

    @property
    def y_H(self) -> Optional[jqt.Qarray]:
        """
        Logical Y hamiltonian.
        """
        return None

    @property
    def z_U(self) -> jqt.Qarray:
        """
        Logical Z unitary gate.
        """
        return self._gen_pauli_U("z")

    @property
    def z_H(self) -> Optional[jqt.Qarray]:
        """
        Logical Z hamiltonian.
        """
        return None


    @property 
    def h_H(self) -> Optional[jqt.Qarray]:
        """
        Logical Hadamard hamiltonian.
        """
        return None

    @property
    def h_U(self) -> jqt.Qarray:
        """
        Logical Hadamard unitary gate.
        """
        return self.basis["+x"] @ self.basis["+z"].dag() + self.basis["-x"] @ self.basis["-z"].dag()

    def _gen_pauli_U(self, basis_state: str) -> jqt.Qarray:
        """
        Generates unitary for Pauli X, Y, Z.

        Args:
            basis_state (str): "x", "y", "z"

        Returns:
            U (jqt.Qarray): Pauli unitary
        """
        H = getattr(self, basis_state + "_H")
        if H is not None:
            return jqt.expm(1.0j * H)

        gate = (
            self.basis["+" + basis_state] @ self.basis["+" + basis_state].dag()
            - self.basis["-" + basis_state] @ self.basis["-" + basis_state].dag()
        )

        return gate

    @property
    def projector(self):
        return (
            self.basis["+z"] @ self.basis["+z"].dag()
            + self.basis["-z"] @ self.basis["-z"].dag()
        )

    @property
    def maximally_mixed_state(self):
        return (1 / 2.0) * self.projector()

    # Plotting
    # ======================================================
    def _prepare_state_plot(self, state):
        """
        Can be overriden.

        E.g. in the case of cavity x transmon system
        return qt.ptrace(state, 0)
        """
        return state

    def plot(self, state, ax=None, qp_type=jqt.WIGNER, **kwargs) -> None:
        if ax is None:
            fig, ax = plt.subplots(1, figsize=(4, 3), dpi=200)
        fig = ax.get_figure()

        if qp_type == jqt.WIGNER:
            vmin = -1
            vmax = 1
        elif qp_type == jqt.QFUNC:
            vmin = 0
            vmax = 1

        w_plt = self._plot_single(state, ax=ax, qp_type=qp_type, **kwargs)

        ax.set_title(qp_type.capitalize() + " Quasi-Probability Dist.")
        ticks = np.linspace(vmin, vmax, 5)
        fig.colorbar(w_plt, ax=ax, ticks=ticks)
        ax.set_xlabel(r"Re$(\alpha)$")
        ax.set_ylabel(r"Im$(\alpha)$")
        fig.tight_layout()

        plt.show()

    def _plot_single(self, state, ax=None, contour=True, qp_type=jqt.WIGNER):
        """
        Assumes state has same dims as initial_state.
        """
        state = self.jqt2qt(state)

        if ax is None:
            _, ax = plt.subplots(1, figsize=(4, 3), dpi=200)

        return jqt.plot_qp(
            state, self.wigner_pts, ax=ax, contour=contour, qp_type=qp_type
        )

    def plot_code_states(self, qp_type: str = jqt.WIGNER, **kwargs):
        """
        Plot |±x⟩, |±y⟩, |±z⟩ code states.

        Args:
            qp_type (str): 
                WIGNER or QFUNC
            
        Return:
            axs: Axes 
        """
        fig, axs = plt.subplots(2, 3, figsize=(12, 6), dpi=200)
        if qp_type == jqt.WIGNER:
            cbar_title = r"$\frac{\pi}{2} W(\alpha)$"
            vmin = -1
            vmax = 1
        elif qp_type == jqt.QFUNC:
            cbar_title = r"$\pi Q(\alpha)$"
            vmin = 0
            vmax = 1

        for i, label in enumerate(["+z", "+x", "+y", "-z", "-x", "-y"]):
            state = self._prepare_state_plot(self.basis[label])
            pos = (i // 3, i % 3)
            ax = axs[pos]
            _, w_plt = self._plot_single(state, ax=ax, qp_type=qp_type, **kwargs)
            ax.set_title(f"|{label}" + r"$\rangle$")
            ax.set_xlabel(r"Re[$\alpha$]")
            ax.set_ylabel(r"Im[$\alpha$]")

        fig.suptitle(self.name)
        fig.tight_layout()
        fig.subplots_adjust(right=0.8, hspace=0.2, wspace=0.2)
        fig.align_xlabels(axs)
        fig.align_ylabels(axs)
        cbar_ax = fig.add_axes([0.85 , 0.15, 0.05, 0.7])

        ticks = np.linspace(vmin, vmax, 5)
        fig.colorbar(w_plt, cax=cbar_ax, ticks=ticks)

        cbar_ax.set_title(cbar_title, pad=20)
        fig.tight_layout()
        plt.show()




""" Resonator."""
from flax import struct
from jax import config

import jax.numpy as jnp

from jaxquantum.core.operators import identity, destroy, create
from jaxquantum.devices.superconducting.flux_base import FluxDevice

config.update("jax_enable_x64", True)

@struct.dataclass
class Resonator(FluxDevice):        
    """
    Resonator Device.
    """
    
    def common_ops(self):
        """ Written in the linear basis. """
        ops = {}
        
        N = self.N_pre_diag
        ops["id"] = identity(N)
        ops["a"] = destroy(N)
        ops["a_dag"] = create(N)
        ops["phi"] = self.phi_zpf()*(ops["a"] + ops["a_dag"])  
        ops["n"] = 1j * self.n_zpf() * (ops["a_dag"] - ops["a"])

        return ops

    def phi_zpf(self):
        """Return Phase ZPF."""
        return (2*self.params["Ec"]/self.params["El"])**(.25)
    
    def n_zpf(self):
        n_zpf = (self.params["El"] / (32.0 * self.params["Ec"])) ** (0.25)
        return n_zpf
    
    def get_linear_ω(self):
        """Get frequency of linear terms."""
        return jnp.sqrt(8*self.params["El"]*self.params["Ec"])
    
    def get_H_linear(self):
        """Return linear terms in H."""
        w = self.get_linear_ω()
        return w*(self.linear_ops["a_dag"]@self.linear_ops["a"] + 1/2)
  
    def get_H_full(self):
        """Return full H in linear basis."""
        return self.get_H_linear()
    
    def potential(self, phi):
        """Return potential energy for a given phi."""
        return 0.5 * self.params["El"] * (2 * jnp.pi * phi) ** 2


"""
Converting between different object types.
"""

from numbers import Number
from jax import config, Array
from qutip import Qobj
from typing import Optional, Union, List
import jax.numpy as jnp
import numpy as np


from jaxquantum.core.qarray import Qarray
from jaxquantum.core.dims import DIMS_TYPE, Qtypes


config.update("jax_enable_x64", True)

# Convert between QuTiP and JAX
# ===============================================================
def qt2jqt(qt_obj, dtype=jnp.complex128):
    """QuTiP state -> Qarray.

    Args:
        qt_obj: QuTiP state.
        dtype: JAX dtype.

    Returns:
        Qarray.
    """
    if isinstance(qt_obj, Qarray) or qt_obj is None:
        return qt_obj
    return Qarray.create(jnp.array(qt_obj.full(), dtype=dtype), dims=qt_obj.dims)


def jqt2qt(jqt_obj):
    """Qarray -> QuTiP state.

    Args:
        jqt_obj: Qarray.
        dims: QuTiP dims.

    Returns:
        QuTiP state.
    """
    if isinstance(jqt_obj, Qobj) or jqt_obj is None:
        return jqt_obj
    
    
    if jqt_obj.is_batched:
        res = []
        for i in range(len(jqt_obj)):
            res.append(jqt2qt(jqt_obj[i]))
        return res
    
    dims = [list(jqt_obj.dims[0]), list(jqt_obj.dims[1])]
    return Qobj(np.array(jqt_obj.data), dims=dims)

def extract_dims(arr: Array, dims: Optional[Union[DIMS_TYPE, List[int]]] = None):
    """Extract dims from a JAX array or Qarray.

    Args:
        arr: JAX array or Qarray.
        dims: Qarray dims.

    Returns:
        Qarray dims.
    """
    if isinstance(dims[0], Number):
        is_op = arr.shape[-2] == arr.shape[-1]
        if is_op:
            dims = [dims, dims]
        else:
            dims = [dims, [1] * len(dims)] # defaults to ket 
    return dims
                 

def jnp2jqt(arr: Array, dims: Optional[Union[DIMS_TYPE, List[int]]] = None):
    """JAX array -> QuTiP state.

    Args:
        jnp_obj: JAX array.
        dims: Qarray dims.

    Returns:
        QuTiP state.
    """
    dims = extract_dims(arr, dims) if dims is not None else None
    return Qarray.create(arr, dims=dims)



"""
Qubit
"""

from typing import Tuple
import warnings

from jaxquantum.codes.base import BosonicQubit
import jaxquantum as jqt

from jax import config
import jax.numpy as jnp
import matplotlib.pyplot as plt
import qutip as qt

config.update("jax_enable_x64", True)


class Qubit(BosonicQubit):
    """
    FockQubit
    """

    def _params_validation(self):
        super()._params_validation()
        self.params["N"] = 2

    def _get_basis_z(self) -> Tuple[jqt.Qarray, jqt.Qarray]:
        """
        Construct basis states |+-x>, |+-y>, |+-z>
        """
        N = int(self.params["N"])
        plus_z = jqt.basis(N, 0)
        minus_z = jqt.basis(N, 1)
        return plus_z, minus_z

    @property
    def x_U(self) -> jqt.Qarray:
        return jqt.sigmax()

    @property
    def y_U(self) -> jqt.Qarray:
        return jqt.sigmay()

    @property
    def z_U(self) -> jqt.Qarray:
        return jqt.sigmaz()

    def plot(self, state, ax=None, qp_type="", **kwargs) -> None:
        state = self.jqt2qt(state)
        with warnings.catch_warnings():
            # TODO: suppressing deprecation warnings, deal with this
            warnings.simplefilter("ignore")
            b = qt.Bloch()
            b.add_states(state)
            b.render()
            b.show()
            plt.tight_layout()
            plt.show()



"""
JAX Utils
"""

from numbers import Number
from typing import Dict

from jax import lax, Array, device_put, config
from jax._src.scipy.special import gammaln
import jax.numpy as jnp
import numpy as np

config.update("jax_enable_x64", True)


def device_put_params(params: Dict, non_device_params=None):
    non_device_params = [] if non_device_params is None else non_device_params
    for param in params:
        if param in non_device_params:
            continue
        if isinstance(params[param], Number) or isinstance(params[param], np.ndarray):
            params[param] = device_put(params[param])
    return params


def comb(N, k):
    """
    NCk

    #TODO: replace with jsp.special.comb once issue is closed:
    https://github.com/google/jax/issues/9709

    Args:
        N: total items
        k: # of items to choose

    Returns:
        NCk: N choose k
    """
    one = 1
    N_plus_1 = lax.add(N, one)
    k_plus_1 = lax.add(k, one)
    return lax.exp(
        lax.sub(
            gammaln(N_plus_1), lax.add(gammaln(k_plus_1), gammaln(lax.sub(N_plus_1, k)))
        )
    )


def complex_to_real_iso_matrix(A): 
    return jnp.block([[jnp.real(A), -jnp.imag(A)], [jnp.imag(A), jnp.real(A)]])


def real_to_complex_iso_matrix(A):
    N = A.shape[0]
    return A[:N//2,:N//2] + 1j*A[N//2:,:N//2]


def complex_to_real_iso_vector(v): 
    return jnp.block([[jnp.real(v)], [jnp.imag(v)]])


def real_to_complex_iso_vector(v): 
    N = v.shape[0]
    return v[:N//2,:] + 1j*v[N//2:,:]


def imag_times_iso_vector(v):
    N = v.shape[0]
    return jnp.block([[-v[N//2:,:]], [v[:N//2,:]]])


def imag_times_iso_matrix(A):
    N = A.shape[0]
    Ar = A[:N//2,:N//2]
    Ai = A[N//2:,:N//2]
    return jnp.block([[-Ai, -Ar],[Ar,-Ai]])

def conj_transpose_iso_matrix(A):
    N = A.shape[0]
    Ar = A[:N//2,:N//2].T
    Ai = A[N//2:,:N//2].T
    return jnp.block([[Ar, Ai],[-Ai,Ar]])

def robust_isscalar(val):
    is_scalar = isinstance(val, Number) or jnp.isscalar(val)
    if isinstance(val, Array):
        is_scalar = is_scalar or (len(val.shape) == 0)
    return is_scalar




""" Transmon."""

from flax import struct
from jax import config

import jax.numpy as jnp
import jax.scipy as jsp

from jaxquantum.devices.superconducting.flux_base import FluxDevice
from jaxquantum.core.operators import identity, destroy, create

config.update("jax_enable_x64", True)


@struct.dataclass
class TruncatedTransmon(FluxDevice):
    """
    Transmon Device.
    """

    def common_ops(self):
        """ Written in the linear basis. """
        
        ops = {}

        N = self.N_pre_diag
        ops["id"] = identity(N)
        ops["a"] = destroy(N)
        ops["a_dag"] = create(N)
        ops["phi"] = self.phi_zpf() * (ops["a"] + ops["a_dag"])
        ops["n"] = 1j * self.n_zpf() * (ops["a_dag"] - ops["a"])
        return ops

    def phi_zpf(self):
        """Return Phase ZPF."""
        return (2 * self.params["Ec"] / self.params["Ej"]) ** (0.25)

    def n_zpf(self):
        """Return Charge ZPF."""
        return (self.params["Ej"] / (32 * self.params["Ec"])) ** (0.25)

    def get_linear_ω(self):
        """Get frequency of linear terms."""
        return jnp.sqrt(8 * self.params["Ec"] * self.params["Ej"])

    def get_H_linear(self):
        """Return linear terms in H."""
        w = self.get_linear_ω()
        return w * self.linear_ops["a_dag"] @ self.linear_ops["a"]

    def get_H_full(self):
        """Return full H in linear basis."""
        cos_phi_op = (
            jsp.linalg.expm(1j * self.linear_ops["phi"])
            + jsp.linalg.expm(-1j * self.linear_ops["phi"])
        ) / 2

        H_nl = -self.params["Ej"] * cos_phi_op - self.params[
            "Ej"
        ] / 2 * jnp.linalg.matrix_power(self.linear_ops["phi"], 2)
        return self.get_H_linear() + H_nl

    def potential(self, phi):
        """Return potential energy for a given phi."""
        return - self.params["Ej"] * jnp.cos(2 * jnp.pi * phi)




""" Base Drive."""

from abc import abstractmethod, ABC
from typing import Dict, Any

from flax import struct
from jax import tree_util, Array
from jax import config
import jax.numpy as jnp

from jaxquantum.core.qarray import Qarray
from jaxquantum.core.conversions import jnp2jqt

config.update("jax_enable_x64", True)


@struct.dataclass
class Drive(ABC):
    N: int = struct.field(pytree_node=False)
    ωd: float
    _label: int = struct.field(pytree_node=False)

    @classmethod
    def create(cls, M_max, ωd, label=0):
        cls.M_max = M_max
        N = 2 * M_max + 1
        return cls(N, ωd, label)

    @property
    def label(self):
        return self.__class__.__name__ + str(self._label)

    @property
    def ops(self):
        return self.common_ops()

    def common_ops(self) -> Dict[str, Qarray]:
        ops = {}

        M_max = self.M_max

        # Construct M = ∑ₘ m|m><m| operator in drive charge basis
        ops["M"] = jnp2jqt(jnp.diag(jnp.arange(-M_max, M_max + 1)))

        # Construct Id = ∑ₘ|m><m| in the drive charge basis
        ops["id"] = jnp2jqt(jnp.identity(2 * M_max + 1))

        # Construct M₊ ≡ exp(iθ) and M₋ ≡ exp(-iθ) operators for drive
        ops["M-"] = jnp2jqt(jnp.eye(2 * M_max + 1, k=1))
        ops["M+"] = jnp2jqt(jnp.eye(2 * M_max + 1, k=-1))

        # Construct cos(θ) ≡ 1/2 * [M₊ + M₋] = 1/2 * ∑ₘ|m+1><m| + h.c
        ops["cos(θ)"] = 0.5 * (ops["M+"] + ops["M-"])

        # Construct sin(θ) ≡ -i/2 * [M₊ - M₋] = -i/2 * ∑ₘ|m+1><m| + h.c
        ops["sin(θ)"] = -0.5j * (ops["M+"] - ops["M-"])

        # Construct more general drive operators cos(kθ) and sin(kθ)
        for k in range(2, M_max + 1):
            ops[f"M_+{k}"] = jnp2jqt(jnp.eye(2 * M_max + 1, k=-k))
            ops[f"M_-{k}"] = jnp2jqt(jnp.eye(2 * M_max + 1, k=k))
            ops[f"cos({k}θ)"] = 0.5 * (ops[f"M_+{k}"] + ops[f"M_-{k}"])
            ops[f"sin({k}θ)"] = -0.5j * (ops[f"M_+{k}"] - ops[f"M_-{k}"])

        return ops

    #############################################################

    def get_H(self):
        """
        Bare "drive" Hamiltonian (ωd * M) in the extended Hilbert space.
        """
        return self.ωd * self.ops["M"]



""" Oscillator gates. """

from jaxquantum.core.operators import displace, basis, destroy, num
from jaxquantum.circuits.gates import Gate
from jax.scipy.special import factorial
import jax.numpy as jnp
from jaxquantum import Qarray

def D(N, alpha):
    return Gate.create(
        N, 
        name="D",
        params={"alpha": alpha},
        gen_U = lambda params: displace(N,params["alpha"]),
        num_modes=1
    )

def CD(N, beta):
    g = basis(2,0)
    e = basis(2,1)
    
    gg = g @ g.dag()
    ee = e @ e.dag()

    return Gate.create(
        [2,N], 
        name="CD",
        params={"beta": beta},
        gen_U = lambda params: (gg ^ displace(N, params["beta"]/2)) + (ee ^ displace(N, -params["beta"]/2)),
        num_modes=2
    )

def _Kraus_Op(N, err_prob, l):
    """" Returns the Kraus Operators for l-photon loss with probability
    err_prob in a Hilbert Space of size N """
    return jnp.sqrt(jnp.power(err_prob, l)/factorial(l)) * (num(N) * jnp.log(jnp.sqrt(1-err_prob))).expm() * destroy(N).powm(l)


def Amp_Damp(N, err_prob, max_l):
    kmap = lambda params: Qarray.from_list([_Kraus_Op(N, err_prob, l) for l in
                                            range(max_l+1)])
    return Gate.create(
        N,
        name="Amp_Damp",
        params={"err_prob": err_prob, "max_l": max_l},
        gen_KM = kmap,
        num_modes=1
    )


""" Fluxonium."""

from flax import struct
from jax import config
import jaxquantum as jqt
import jax.numpy as jnp

from jaxquantum.devices.superconducting.flux_base import FluxDevice
from jaxquantum.devices.base.base import HamiltonianTypes

config.update("jax_enable_x64", True)


@struct.dataclass
class Fluxonium(FluxDevice):
    """
    Fluxonium Device.
    """

    def common_ops(self):
        """ Written in the linear basis. """
        ops = {}

        N = self.N_pre_diag
        ops["id"] = jqt.identity(N)
        ops["a"] = jqt.destroy(N)
        ops["a_dag"] = jqt.create(N)
        ops["phi"] = self.phi_zpf() * (ops["a"] + ops["a_dag"])
        ops["n"] = 1j * self.n_zpf() * (ops["a_dag"] - ops["a"])

        ops["cos(φ/2)"] = jqt.cosm(ops["phi"] / 2)
        ops["sin(φ/2)"] = jqt.sinm(ops["phi"] / 2)

        return ops

    def n_zpf(self):
        n_zpf = (self.params["El"] / (32.0 * self.params["Ec"])) ** (0.25)
        return n_zpf

    def phi_zpf(self):
        """Return Phase ZPF."""
        return (2 * self.params["Ec"] / self.params["El"]) ** (0.25)

    def get_linear_ω(self):
        """Get frequency of linear terms."""
        return jnp.sqrt(8 * self.params["Ec"] * self.params["El"])

    def get_H_linear(self):
        """Return linear terms in H."""
        w = self.get_linear_ω()
        return w * (
            self.linear_ops["a_dag"] @ self.linear_ops["a"]
            + 0.5 * self.linear_ops["id"]
        )

    def get_H_full(self):
        """Return full H in linear basis."""
        
        phi_op = self.linear_ops["phi"]
        return self.get_H_linear() + self.get_H_nonlinear(phi_op)

    def get_H_nonlinear(self, phi_op):
        op_cos_phi = jqt.cosm(phi_op)
        op_sin_phi = jqt.sinm(phi_op)

        phi_ext = self.params["phi_ext"]
        Hcos = op_cos_phi * jnp.cos(2.0 * jnp.pi * phi_ext) + op_sin_phi * jnp.sin(
            2.0 * jnp.pi * phi_ext
        )
        H_nl = - self.params["Ej"] * Hcos
        return H_nl

    def potential(self, phi):
        """Return potential energy for a given phi."""
        phi_ext = self.params["phi_ext"]
        V_linear = 0.5 * self.params["El"] * (2 * jnp.pi * phi) ** 2
    
        if self.hamiltonian == HamiltonianTypes.linear:
            return V_linear
        
        V_nonlinear = -self.params["Ej"] * jnp.cos(2.0 * jnp.pi * (phi - phi_ext))
        if self.hamiltonian == HamiltonianTypes.full:
            return V_linear + V_nonlinear


""" qubit gates. """

from jaxquantum.core.operators import sigmax, sigmay, sigmaz, basis, hadamard, qubit_rotation
from jaxquantum.circuits.gates import Gate
from jaxquantum.core.qarray import Qarray
import jax.numpy as jnp

def X():
    return Gate.create(
        2, 
        name="X",
        gen_U = lambda params: sigmax(),
        num_modes=1
    )

def Y():
    return Gate.create(
        2, 
        name="Y",
        gen_U = lambda params: sigmay(),
        num_modes=1
    )

def Z():
    return Gate.create(
        2, 
        name="Z",
        gen_U = lambda params: sigmaz(),
        num_modes=1
    )

def H():
    return Gate.create(
        2, 
        name="H",
        gen_U = lambda params: hadamard(),
        num_modes=1
    )

def Rx(theta):
    return Gate.create(
        2, 
        name="Rx",
        params={"theta": theta},
        gen_U = lambda params: qubit_rotation(params["theta"], 1, 0, 0),
        num_modes=1
    )

def Ry(theta):
    return Gate.create(
        2, 
        name="Ry",
        params={"theta": theta},
        gen_U = lambda params: qubit_rotation(params["theta"], 0, 1, 0),
        num_modes=1
    )

def Rz(theta):
    return Gate.create(
        2, 
        name="Rz",
        params={"theta": theta},
        gen_U = lambda params: qubit_rotation(params["theta"], 0, 0, 1),
        num_modes=1
    )

def MZ():

    g = basis(2,0)
    e = basis(2,1)

    gg = g @ g.dag()
    ee = e @ e.dag()

    kmap = Qarray.from_list([gg, ee])

    return Gate.create(
        2, 
        name="MZ",
        gen_KM = lambda params: kmap,
        num_modes=1
    )

def MX():

    g = basis(2,0)
    e = basis(2,1)

    plus = (g + e).unit()
    minus = (g - e).unit()

    pp = plus @ plus.dag()
    mm = minus @ minus.dag()

    kmap = Qarray.from_list([pp, mm])

    return Gate.create(
        2, 
        name="MX",
        gen_KM = lambda params: kmap,
        num_modes=1
    )

def MX_plus():

    g = basis(2,0)
    e = basis(2,1)
    plus = (g + e).unit()
    pp = plus @ plus.dag()
    kmap = Qarray.from_list([2*pp])

    return Gate.create(
        2, 
        name="MXplus",
        gen_KM = lambda params: kmap,
        num_modes=1
    )

def MZ_plus():
    g = basis(2,0)
    plus = g
    pp = plus @ plus.dag()
    kmap = Qarray.from_list([2*pp])

    return Gate.create(
        2, 
        name="MZplus",
        gen_KM = lambda params: kmap,
        num_modes=1
    )

def Reset():
    g = basis(2,0)
    e = basis(2,1)

    gg = g @ g.dag()
    ge = g @ e.dag()

    kmap = Qarray.from_list([gg, ge])
    return Gate.create(
        2,
        name="Reset",
        gen_KM = lambda params: kmap,
        num_modes=1
    )

def IP_Reset(p_eg, p_ee):
    g = basis(2, 0)
    e = basis(2, 1)

    gg = g @ g.dag()
    ge = g @ e.dag()
    eg = e @ g.dag()
    ee = e @ e.dag()

    k_0 = jnp.sqrt(1-p_eg) * gg + jnp.sqrt(p_eg) * eg
    k_1 = jnp.sqrt(p_ee) * ee + jnp.sqrt(1-p_ee) * ge

    kmap = Qarray.from_list([k_0, k_1])

    return Gate.create(
        2,
        name="IP_Reset",
        params={"p_eg": p_eg, "p_ge": p_ee},
        gen_KM=lambda params: kmap,
        num_modes=1
    )

def CX():
    g = basis(2,0)
    e = basis(2,1)
    
    gg = g @ g.dag()
    ee = e @ e.dag()

    op =  (gg ^ identity(2)) + (ee ^ sigmax())
    
    return Gate.create(
        [2,2],
        name="CX",
        gen_U = lambda params: op,
        num_modes=2
    )



"""Quantum Tooling"""

from .operators import *
from .conversions import *
from .visualization import *
from .solvers import *
from .qarray import *
from .settings import SETTINGS
from .dims import *
from .measurements import *


import pytest

from jax import jit
import sys
import os

# Add the jaxquantum directory to the sys.path
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import jaxquantum as jqt
import jax.numpy as jnp


# sesolve ====
def test_sesolve():
    omega_q = 5.0 #GHzz
    Omega = .1
    g_state = jqt.basis(2,0) ^ jqt.basis(2,0)
    g_state_dm = g_state.to_dm()

    ts = jnp.linspace(0,5*jnp.pi/Omega,101)

    sz0 = jqt.sigmaz() ^ jqt.identity(N=2)

    @jit
    def Ht(t):
        H0 = omega_q/2.0*((jqt.sigmaz()^jqt.identity(N=2)) + (jqt.identity(N=2)^jqt.sigmaz()))
        H1 = Omega*jnp.cos((omega_q)*t)*((jqt.sigmax()^jqt.identity(N=2)) + (jqt.identity(N=2)^jqt.sigmax()))
        return H0 + H1

    states = jqt.sesolve(Ht, g_state, ts) 
    szt = jnp.real(jqt.overlap(states, sz0))

    test_time = ts[34]
    sim_szt =  szt[34]
    cal_szt = jnp.cos(Omega*test_time)
    assert jnp.isclose(sim_szt, cal_szt, atol=1e-4), f"Expected {cal_szt}, got {sim_szt}"

def test_sesolve_batch():
    omega_q = 5.0 #GHzz
    Omega = jnp.array([.1,.2])
    g_state = jqt.basis(2,0) ^ jqt.basis(2,0)
    g_state_dm = g_state.to_dm()

    ts = jnp.linspace(0,5*jnp.pi/Omega[0],101)

    sz0 = jqt.sigmaz() ^ jqt.identity(N=2)

    @jit
    def Ht(t):
        H0 = omega_q/2.0*((jqt.sigmaz()^jqt.identity(N=2)) + (jqt.identity(N=2)^jqt.sigmaz()))
        H1 = Omega*jnp.cos((omega_q)*t)*((jqt.sigmax()^jqt.identity(N=2)) + (jqt.identity(N=2)^jqt.sigmax()))
        return H0 + H1

    states = jqt.sesolve(Ht, g_state, ts) 
    szt = jnp.real(jqt.overlap(states, sz0))

    for j in range(2):
        test_time = ts[34]
        sim_szt =  szt[34,j]
        cal_szt = jnp.cos(Omega[j]*test_time)
        assert jnp.isclose(sim_szt, cal_szt, atol=1e-3), f"Expected {cal_szt}, got {sim_szt}"


def test_sesolve_edge_cases():
    # constant H0

    omega_q = 5.0 #GHzz
    Omega = .1
    g_state = jqt.basis(2,0) ^ jqt.basis(2,0)
    ts = jnp.linspace(0,5*jnp.pi/Omega,101)
    sz0 = jqt.sigmaz() ^ jqt.identity(N=2)
    H0 = Omega/2*jqt.sigmax() ^ jqt.identity(N=2)
    states = jqt.sesolve(H0, g_state, ts)

    szt = jnp.real(jqt.overlap(states, sz0))

    test_time = ts[50]
    test_szt = szt[50]
    cal_szt = jnp.cos(Omega*test_time)
    assert jnp.isclose(test_szt, cal_szt, atol=1e-5)

    # valueerror if initial state is dm
    with pytest.raises(ValueError):
        jqt.sesolve(H0, g_state.to_dm(), ts)

# ====

# mesolve ====

def test_mesolve_batch():
    N = 100

    omega_a = 2.0*jnp.pi*5.0
    kappa = 2*jnp.pi*jnp.array([1,2]) # Batching to explore two different kappa values!
    initial_state = jqt.displace(N, 0.1) @ jqt.basis(N,0)
    initial_state_dm = initial_state.to_dm()
    ts = jnp.linspace(0, 4*2*jnp.pi/omega_a, 101)

    a = jqt.destroy(N)
    n = a.dag() @ a

    c_ops = jqt.Qarray.from_list([jnp.sqrt(kappa)*a])

    @jit
    def Ht(t):
        H0 = omega_a*n
        return H0

    solver_options = jqt.SolverOptions.create(progress_meter=True)
    states = jqt.mesolve(Ht, initial_state_dm, ts, c_ops=c_ops, solver_options=solver_options) 
    nt = jnp.real(jqt.overlap(n, states))
    a_real = jnp.real(jqt.overlap(a, states))
    a_imag = jnp.imag(jqt.overlap(a, states))

    for j in range(2):
        test_time = ts[50]
        test_nt = nt[50,j]
        expected_nt = jnp.exp(-kappa[j]*test_time) * jnp.abs(jqt.overlap(n, initial_state_dm))  # Expectation value of n at time t
        assert jnp.isclose(test_nt, expected_nt, atol=1e-8), f"Expected {expected_nt}, got {test_nt}"

def test_mesolve():
    N = 100

    omega_a = 2.0*jnp.pi*5.0
    kappa = 2*jnp.pi*1.0 # Batching to explore two different kappa values!
    initial_state = jqt.displace(N, 0.1) @ jqt.basis(N,0)
    initial_state_dm = initial_state.to_dm()
    ts = jnp.linspace(0, 4*2*jnp.pi/omega_a, 101)

    a = jqt.destroy(N)
    n = a.dag() @ a

    c_ops = jqt.Qarray.from_list([jnp.sqrt(kappa)*a])

    @jit
    def Ht(t):
        H0 = omega_a*n
        return H0

    solver_options = jqt.SolverOptions.create(progress_meter=True)
    states = jqt.mesolve(Ht, initial_state_dm, ts, c_ops=c_ops, solver_options=solver_options) 
    nt = jnp.real(jqt.overlap(n, states))
    a_real = jnp.real(jqt.overlap(a, states))
    a_imag = jnp.imag(jqt.overlap(a, states))

    test_time = ts[50]
    test_nt = nt[50]
    expected_nt = jnp.exp(-kappa*test_time) * jnp.abs(jqt.overlap(n, initial_state_dm))  # Expectation value of n at time t
    assert jnp.isclose(test_nt, expected_nt, atol=1e-8), f"Expected {expected_nt}, got {test_nt}"


def test_mesolve_edge_cases():
    # constant H0

    omega_q = 5.0 #GHzz
    Omega = .1
    g_state = jqt.basis(2,0) ^ jqt.basis(2,0)
    ts = jnp.linspace(0,5*jnp.pi/Omega,101)
    sz0 = jqt.sigmaz() ^ jqt.identity(N=2)
    H0 = Omega/2*jqt.sigmax() ^ jqt.identity(N=2)
    c_ops = jqt.Qarray.from_list([])
    states = jqt.mesolve(H0, g_state, ts, c_ops=c_ops)

    szt = jnp.real(jqt.overlap(states, sz0))

    test_time = ts[50]
    test_szt = szt[50]
    cal_szt = jnp.cos(Omega*test_time)
    assert jnp.isclose(test_szt, cal_szt, atol=1e-5)


# ====




""" Gates. """

import functools
from flax import struct
from enum import Enum
from jax import Array, config
from typing import List, Dict, Any, Optional, Callable, Union
from math import prod
from copy import deepcopy
from numbers import Number
import jax.numpy as jnp
import jax.scipy as jsp


from jaxquantum.core.settings import SETTINGS
from jaxquantum.core.qarray import Qarray

config.update("jax_enable_x64", True)


@struct.dataclass
class Gate:
    dims: List[int] = struct.field(pytree_node=False)
    _U: Optional[Array] # Unitary
    _H: Optional[Array] # Hamiltonian
    _KM: Optional[Qarray] # Kraus map
    _params: Dict[str, Any]
    _ts: Array
    _name: str = struct.field(pytree_node=False)
    num_modes: int = struct.field(pytree_node=False)
    
    @classmethod
    def create(
        cls,
        dims: Union[int, List[int]],
        name: str = "Gate",
        params: Optional[Dict[str, Any]] = None,
        ts: Optional[Array] = None,
        gen_U: Optional[Callable[[Dict[str, Any]], Qarray]] = None,
        gen_H: Optional[Callable[[Dict[str, Any]], Qarray]] = None,
        gen_KM: Optional[Callable[[Dict[str, Any]], List[Qarray]]] = None,
        num_modes: int = 1,
    ):
        """ Create a gate. 
        
        Args:
            dims: Dimensions of the gate.
            name: Name of the gate.
            params: Parameters of the gate.
            ts: Times of the gate.
            gen_U: Function to generate the unitary of the gate.
            gen_H: Function to generate the Hamiltonian of the gate.
            gen_KM: Function to generate the Kraus map of the gate.
            num_modes: Number of modes of the gate.
        """

        # TODO: add params to device?

        if isinstance(dims, int):
            dims = [dims]
        
        assert len(dims) == num_modes, "Number of dimensions must match number of modes."
        

        # Unitary
        _U = gen_U(params) if gen_U is not None else None 
        _H = gen_H(params) if gen_H is not None else None 

        if gen_KM is not None:
            _KM = gen_KM(params)
        elif _U is not None:
            _KM = Qarray.from_list([_U])

        return Gate(
            dims = dims,
            _U = _U,
            _H = _H,
            _KM = _KM,
            _params = params if params is not None else {},
            _ts=ts if ts is not None else jnp.array([]),
            _name = name,
            num_modes = num_modes
        )

    def __str__(self):
        return self._name

    def __repr__(self):
        return self._name

    @property
    def U(self):
        return self._U

    @property
    def H(self):
        return self._H

    @property
    def KM(self):
        return self._KM


"""
Visualization utils.
"""

import qutip as qt
import numpy as np
import matplotlib.pyplot as plt

from jaxquantum.core.conversions import jqt2qt

WIGNER = "wigner"
QFUNC = "qfunc"


def plot_qp(state, pts, ax=None, contour=True, qp_type=WIGNER, cbar_label="", axis_scale_factor=1, plot_cbar=True):
    """Plot quasi-probability distribution.

    TODO: decouple this from qutip.

    Args:
        state: statevector
        pts: points to evaluate quasi-probability distribution on
        dim: dimensions of state
        ax: matplotlib axis to plot on
        contour: make the plot use contouring
        qp_type: type of quasi probability distribution ("wigner", "qfunc")

    Returns:
        axis on which the plot was plotted.
    """
    pts = np.array(pts)
    state = jqt2qt(state)
    if ax is None:
        _, ax = plt.subplots(1, figsize=(4, 3), dpi=200)
    # fig = ax.get_figure()

    if qp_type == WIGNER:
        vmin = -1
        vmax = 1
        scale = np.pi / 2
        cmap = "seismic"
    elif qp_type == QFUNC:
        vmin = 0
        vmax = 1
        scale = np.pi
        cmap = "jet"

    QP = scale * getattr(qt, qp_type)(state, pts, pts, g=2)

    pts = pts * axis_scale_factor

    if contour:
        im = ax.contourf(
            pts,
            pts,
            QP,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            levels=np.linspace(vmin, vmax, 101),
        )
    else:
        im = ax.pcolormesh(
            pts,
            pts,
            QP,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )
    ax.axhline(0, linestyle="-", color="black", alpha=0.7)
    ax.axvline(0, linestyle="-", color="black", alpha=0.7)
    ax.grid()
    ax.set_aspect("equal", adjustable="box")

    if plot_cbar:
        cbar = plt.colorbar(im, ax=ax, orientation="vertical", ticks=np.linspace(-1, 1, 11))
        cbar.ax.set_title(cbar_label)

    ax.set_xlabel(r"Re[$\alpha$]")
    ax.set_ylabel(r"Im[$\alpha$]")

    fig = ax.get_figure()
    fig.tight_layout()
    return ax, im


plot_wigner = lambda state, pts, ax=None, contour=True, **kwargs: plot_qp(
    state, pts, ax=ax, contour=contour, qp_type=WIGNER, cbar_label=r"$\mathcal{W}(\alpha)$", **kwargs
)

plot_qfunc = lambda state, pts, ax=None, contour=True: plot_qp(
    state, pts, ax=ax, contour=contour, qp_type=QFUNC, cbar_label=r"$\mathcal{Q}(\alpha)$", **kwargs
)



""" Utility functions """

from scipy.special import pbdv
from scipy import constants
from jax import vmap, grad

import jax.scipy as jsp
import jax.numpy as jnp

def factorial_approx(n):
    return jsp.special.gamma(n+1)


# physics utils

# ----------------



def harm_osc_wavefunction(
    n, x, l_osc
):
    r"""
    Taken from scqubits... not jit-able
    
    For given quantum number n=0,1,2,... return the value of the harmonic
    oscillator wave function :math:`\psi_n(x) = N H_n(x/l_{osc}) \exp(-x^2/2l_\text{
    osc})`, N being the proper normalization factor.

    Directly uses `scipy.special.pbdv` (implementation of the parabolic cylinder
    function) to mitigate numerical stability issues with the more commonly used
    expression in terms of a Gaussian and a Hermite polynomial factor.

    Parameters
    ----------
    n:
        index of wave function, n=0 is ground state
    x:
        coordinate(s) where wave function is evaluated
    l_osc:
        oscillator length, defined via <0|x^2|0> = l_osc^2/2

    Returns
    -------
        value of harmonic oscillator wave function
    """
    x = 2 * jnp.pi * x
    result = pbdv(n, jnp.sqrt(2.0) * x / l_osc) [0]
    result = result / jnp.sqrt(
        l_osc * jnp.sqrt(jnp.pi) *  factorial_approx(n)
    )
    return result



def calculate_lambda_over_four_resonator_zpf(freq, impedance):
    expected_Z0 = impedance  # Ohms
    expected_E_L_over_E_C = (1 / (4 * expected_Z0)) ** 2 * (
        constants.h**2 / (8 * constants.e**4)
    )
    desired_E_C = jnp.sqrt(freq**2 / expected_E_L_over_E_C / 8)
    desired_E_L = freq**2 / desired_E_C / 8
    storage_q_zpf = (1 / 32 * desired_E_L / desired_E_C) ** (1 / 4)
    return storage_q_zpf, desired_E_C, desired_E_L


""" Circuit simulation methods. """

import functools
from flax import struct
from enum import Enum
from jax import Array, config
from typing import List, Union
from math import prod
from copy import deepcopy
from numbers import Number
import jax.numpy as jnp
import jax.scipy as jsp


from jaxquantum.core.settings import SETTINGS
from jaxquantum.core.qarray import Qarray, ket2dm
from jaxquantum.circuits.circuits import Circuit, Layer
from jaxquantum.circuits.constants import SimulateMode

config.update("jax_enable_x64", True)


@struct.dataclass
class Results:
    results: List[Qarray] = struct.field(pytree_node=False)

    @classmethod
    def create(
        cls,
        results: List[Qarray]
    ):
        return Results(
            results = results
        )

    def __getitem__(self, j: int):
        return self.results[j]

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return str(self.results)

    def append(self, result: Qarray):
        self.results.append(result)

    def __getitem__(self, j: int):
        return self.results[j]

    def __len__(self):
        return len(self.results)

def simulate(
    circuit: Circuit,
    initial_state: Qarray,
    mode: SimulateMode = SimulateMode.UNITARY
) -> Results:
    """
    Simulates the evolution of a quantum state through a given quantum circuit.

    Args:
        circuit (Circuit): The quantum circuit to simulate. The circuit is composed of layers, 
                           each of which can generate unitary or Kraus operators.
        initial_state (Qarray): The initial quantum state to be evolved. This can be a state vector 
                                or a density matrix.
        mode (SimulateMode, optional): The mode of simulation. It can be either SimulateMode.UNITARY 
                                       for unitary evolution or SimulateMode.KRAUS for Kraus operator 
                                       evolution. Defaults to SimulateMode.UNITARY.

    Returns:
        Results: An object containing the results of the simulation, which includes the quantum states 
                 at each step of the circuit.
    """

    results = Results.create([])
    state = initial_state
    results.append(Qarray.from_list([state]))

    for layer in circuit.layers:
        result = simulate_layer(layer, state, mode=mode)
        results.append(result)
        state = result[-1]

    return results

def simulate_layer(layer: Layer, initial_state: Qarray, mode: SimulateMode = SimulateMode.UNITARY) -> Qarray:
    """
    Simulates the evolution of a quantum state through a given layer.

    Args:
        layer (Layer): The layer through which the quantum state evolves. 
                       This layer should have methods to generate unitary (gen_U) 
                       and Kraus (gen_KM) operators.
        initial_state (Qarray): The initial quantum state to be evolved. 
                                This can be a state vector or a density matrix.
        mode (SimulateMode, optional): The mode of simulation. It can be either 
                                       SimulateMode.UNITARY for unitary evolution 
                                       or SimulateMode.KRAUS for Kraus operator evolution
                                       or SimulateMode.DEFAULT to use the default simulate mode in the layer.
                                       Defaults to SimulateMode.UNITARY.
    Returns:
        Qarray: The result of the simulation containing the evolved quantum state.
    """

    state = initial_state 

    if mode == SimulateMode.DEFAULT:
        mode = layer._default_simulate_mode

    if mode == SimulateMode.UNITARY:
        U = layer.gen_U()
        if state.is_dm():
            state = U @ state @ U.dag()
        else:
            state = U @ state 
        
        result = Qarray.from_list([state])

    elif mode == SimulateMode.KRAUS:
        KM = layer.gen_KM()
        
        state = ket2dm(state)
        state = (KM @ state @ KM.dag()).collapse()

        # new_state = 0
        # for op_j in range(len(KM)):
        #     op = KM[op_j]
        #     new_state += op @ state @ op.dag()
        # state = new_state

        result = Qarray.from_list([state])
    
    return result


        
    



""" Sweeping tools. """

from copy import deepcopy
import itertools
from tempfile import NamedTemporaryFile
import jax.numpy as jnp
from tqdm import tqdm
import os



def run_sweep(params, sweep_params, metrics_func, fixed_kwargs=None, data=None, is_parallel=False, save_file=None, data_save_mode="end", return_errors=False):
    """ Run a sweep over a single parameter, or multiple parameters.

    Args:
        params (dict): The base parameters to sweep over.
        sweep_params (dict): The parameters to sweep over.
            key: The parameter name.
            value: The list of values to sweep over.
        metrics_func (function): The function to evaluate the metrics.
        fixed_params (dict, optional): The fixed parameters to send into metrics_func. Defaults to None.
        data (dict, optional): The data to append to. Defaults to None.
        is_parallel (bool, optional): Whether to sweep through the sweep_params lists in parallel or through their cartesian product. Defaults to False.
        save_file (str, optional): The file to save the data to. Defaults to None, in which case data is saved to a temporary file, which will be deleted upon closing (e.g. during garbage collection).
        data_save_mode (str, optional): The mode to save the data. Defaults to None. 
            Options are: 
                "no" - don't save data 
                "end" - save data at the end of the sweep
                "during" - save data during and at the end of the sweep
    Returns:
        dict: The data after the sweep.
    """
    if data is None:
        data = {}
        run = -1
    else:
        run = max(data.keys())

    assert data_save_mode in ["no", "end", "during"], "Invalid data_save_mode."

    if data_save_mode in ["during", "end"]:
        if isinstance(save_file, str):
            print("Saving data to: ", save_file)
            dirname = os.path.dirname(save_file)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
        else:
            save_file = save_file or NamedTemporaryFile()
            print("Saving data to a temporary file: ", save_file.name)
        

    fixed_kwargs = fixed_kwargs or {}

    if is_parallel:
        sweep_length = len(list(sweep_params.values())[0])
        assert [len(vals) == sweep_length for vals in sweep_params.values()], "Parallel sweep parameters must have the same length."

        errors = []
        try:
            for j in tqdm(range(sweep_length)):
                run += 1
                data[run] = {}
                data[run]["params"] = deepcopy(params)
                sweep_point_info = {
                    "labels": [],
                    "values": [],
                    "indices": [],
                }
                for key, vals in sweep_params.items():
                    data[run]["params"][key] = vals[j]
                    sweep_point_info["labels"].append(key)
                    sweep_point_info["values"].append(vals[j])
                    sweep_point_info["indices"].append(j)
                data[run]["results"] = metrics_func(data[run]["params"], **fixed_kwargs)
                data[run]["sweep_point_info"] = sweep_point_info
                if data_save_mode == "during":
                    jnp.savez(save_file, data=data, sweep_params=sweep_params, params=params)
        except Exception as e:
            errors.append(str(e))
            print("Error during run: ", errors[-1])
        
        try:
            if data_save_mode in ["during", "end"]:
                jnp.savez(save_file, data=data, sweep_params=sweep_params, params=params, error=None)
        except Exception as e:
            errors.append(str(e))
            print("Error during saving: ", errors[-1])

        if return_errors:
            return data, errors
        else:
            return data
    else:
        # Product Sweep
        sweep_points = list(itertools.product(*list(sweep_params.values())))
        sweep_points_indxs = list(itertools.product(*[list(range(len(vals))) for vals in list(sweep_params.values())]))
        sweep_point_labels = list(sweep_params.keys())

        errors = []

        try:
            with tqdm(total=len(sweep_points)) as pbar:
                for j, sweep_point in enumerate(sweep_points):
                    run += 1
                    data[run] = {}
                    data[run]["params"] = deepcopy(params)
                    sweep_point_info = {
                        "labels": [],
                        "values": [],
                        "indices": [],
                    }
                    for i, key in enumerate(sweep_point_labels):
                        data[run]["params"][key] = sweep_point[i]
                        sweep_point_info["labels"].append(key)
                        sweep_point_info["values"].append(sweep_point[i])
                        sweep_point_info["indices"].append(sweep_points_indxs[j][i])
                    data[run]["results"] = metrics_func(data[run]["params"], **fixed_kwargs)
                    data[run]["sweep_point_info"] = sweep_point_info
                    pbar.update(1)
                    if data_save_mode == "during":
                        jnp.savez(save_file, data=data, sweep_params=sweep_params, params=params)
        except Exception as e:
            errors.append(str(e))
            print("Error during run: ", errors[-1])
            
        try:
            if data_save_mode in ["during", "end"]:
                jnp.savez(save_file, data=data, sweep_params=sweep_params, params=params)
        except Exception as e:
            errors.append(str(e))
            print("Error during saving: ", errors[-1])

        if return_errors:
            return data, errors
        else:
            return data


""" Transmon."""

from flax import struct
from jax import config

import jax.numpy as jnp
import jax.scipy as jsp

from jaxquantum.devices.base.base import BasisTypes, HamiltonianTypes
from jaxquantum.devices.superconducting.flux_base import FluxDevice
from jaxquantum.core.operators import identity, destroy, create
from jaxquantum.core.conversions import jnp2jqt

config.update("jax_enable_x64", True)


@struct.dataclass
class Transmon(FluxDevice):
    """
    Transmon Device.
    """
    DEFAULT_BASIS = BasisTypes.charge
    DEFAULT_HAMILTONIAN = HamiltonianTypes.full

    @classmethod
    def param_validation(cls, N, N_pre_diag, params, hamiltonian, basis):
        """ This can be overridden by subclasses."""
        if hamiltonian == HamiltonianTypes.linear:
            assert basis == BasisTypes.fock, "Linear Hamiltonian only works with Fock basis."
        elif hamiltonian == HamiltonianTypes.truncated:
            assert basis == BasisTypes.fock, "Truncated Hamiltonian only works with Fock basis."
        elif hamiltonian == HamiltonianTypes.full:
            assert basis in [BasisTypes.charge, BasisTypes.single_charge], "Full Hamiltonian only works with Cooper pair charge or single-electron charge bases."
        
        # Set the gate offset charge to zero if not provided
        if "ng" not in params:
            params["ng"] = 0.0
        
        assert (N_pre_diag - 1) % 2 == 0, "N_pre_diag must be odd."

    def common_ops(self):
        """ Written in the specified basis. """
        
        ops = {}

        N = self.N_pre_diag

        if self.basis == BasisTypes.fock:
            ops["id"] = identity(N)
            ops["a"] = destroy(N)
            ops["a_dag"] = create(N)
            ops["phi"] = self.phi_zpf() * (ops["a"] + ops["a_dag"])
            ops["n"] = 1j * self.n_zpf() * (ops["a_dag"] - ops["a"])

        elif self.basis == BasisTypes.charge:
            """
            Here H = 4 * Ec (n - ng)² - Ej cos(φ) in the Cooper pair charge basis. 
            """
            ops["id"] = identity(N)
            ops["cos(φ)"] = 0.5*(jnp2jqt(jnp.eye(N,k=1) + jnp.eye(N,k=-1)))
            ops["sin(φ)"] = 0.5j*(jnp2jqt(jnp.eye(N,k=1) - jnp.eye(N,k=-1)))
            n_max = (N - 1) // 2
            ops["n"] = jnp2jqt(jnp.diag(jnp.arange(-n_max, n_max + 1)))

            n_minus_ng_array = jnp.arange(-n_max, n_max + 1) - self.params["ng"] * jnp.ones(N)
            ops["H_charge"] = jnp2jqt(jnp.diag(4 * self.params["Ec"] * n_minus_ng_array**2))

        elif self.basis == BasisTypes.single_charge:
            """
            Here H = Ec (n - 2ng)² - Ej cos(φ) in the single-electron charge basis. Using Eq. (5.36) of Kyle Serniak's
            thesis, we have H = Ec ∑ₙ(n - 2*ng) |n⟩⟨n| - Ej/2 * ∑ₙ|n⟩⟨n+2| + h.c where n counts the number of electrons, 
            not Cooper pairs. Note, we use 2ng instead of ng to match the gate offset charge convention of the transmon 
            (as done in Kyle's thesis).
            """
            n_max = (N - 1) // 2

            ops["id"] = identity(N)
            ops["cos(φ)"] = 0.5*(jnp2jqt(jnp.eye(N,k=2) + jnp.eye(N,k=-2)))
            ops["sin(φ)"] = 0.5j*(jnp2jqt(jnp.eye(N,k=2) - jnp.eye(N,k=-2)))
            ops["cos(φ/2)"] = 0.5*(jnp2jqt(jnp.eye(N,k=1) + jnp.eye(N,k=-1)))
            ops["sin(φ/2)"] = 0.5j*(jnp2jqt(jnp.eye(N,k=1) - jnp.eye(N,k=-1)))
            ops["n"] = jnp2jqt(jnp.diag(jnp.arange(-n_max, n_max + 1)))

            n_minus_ng_array = jnp.arange(-n_max, n_max + 1) - 2 * self.params["ng"] * jnp.ones(N)
            ops["H_charge"] = jnp2jqt(jnp.diag(self.params["Ec"] * n_minus_ng_array**2))

        return ops

    @property
    def Ej(self):
        return self.params["Ej"]

    def phi_zpf(self):
        """Return Phase ZPF."""
        return (2 * self.params["Ec"] / self.Ej) ** (0.25)

    def n_zpf(self):
        """Return Charge ZPF."""
        return (self.Ej / (32 * self.params["Ec"])) ** (0.25)

    def get_linear_ω(self):
        """Get frequency of linear terms."""
        return jnp.sqrt(8 * self.params["Ec"] * self.Ej)

    def get_H_linear(self):
        """Return linear terms in H."""
        w = self.get_linear_ω()
        return w * self.original_ops["a_dag"] @ self.original_ops["a"]

    def get_H_full(self):
        """Return full H in specified basis."""
        return self.original_ops["H_charge"] - self.Ej * self.original_ops["cos(φ)"]
    
    def get_H_truncated(self):
        """Return truncated H in specified basis."""
        phi_op = self.original_ops["phi"]  
        fourth_order_term =  - (1/24) * self.Ej * phi_op @ phi_op @ phi_op @ phi_op 
        sixth_order_term = (1/720) * self.Ej * phi_op @ phi_op @ phi_op @ phi_op @ phi_op @ phi_op
        return self.get_H_linear() + fourth_order_term + sixth_order_term
    
    def _get_H_in_original_basis(self):
        """ This returns the Hamiltonian in the original specified basis. This can be overridden by subclasses."""

        if self.hamiltonian == HamiltonianTypes.linear:
            return self.get_H_linear()
        elif self.hamiltonian == HamiltonianTypes.full:
            return self.get_H_full()
        elif self.hamiltonian == HamiltonianTypes.truncated:
            return self.get_H_truncated()

    def potential(self, phi):
        """Return potential energy for a given phi."""
        if self.hamiltonian == HamiltonianTypes.linear:
            return 0.5 * self.Ej * (2 * jnp.pi * phi) ** 2
        elif self.hamiltonian == HamiltonianTypes.full:
            return - self.Ej * jnp.cos(2 * jnp.pi * phi)
        elif self.hamiltonian == HamiltonianTypes.truncated:
            phi_scaled = 2 * jnp.pi * phi
            second_order = 0.5 * self.Ej * phi_scaled ** 2
            fourth_order =  - (1/24) * self.Ej * phi_scaled ** 4
            sixth_order = (1/720) * self.Ej * phi_scaled ** 6
            return second_order + fourth_order + sixth_order

    def calculate_wavefunctions(self, phi_vals):
        """Calculate wavefunctions at phi_exts."""

        if self.basis == BasisTypes.fock:
            return super().calculate_wavefunctions(phi_vals)
        elif self.basis == BasisTypes.single_charge:
            raise NotImplementedError("Wavefunctions for single charge basis not yet implemented.")
        elif self.basis == BasisTypes.charge:
            phi_vals = jnp.array(phi_vals)

            n_labels = jnp.diag(self.original_ops["n"].data)

            wavefunctions = []
            for nj in range(self.N_pre_diag):
                wavefunction = []
                for phi in phi_vals:
                    wavefunction.append(
                        (1j ** nj / jnp.sqrt(2*jnp.pi)) * jnp.sum(
                            self.eig_systems["vecs"][:,nj] * jnp.exp(1j * phi * n_labels)
                        )
                    )
                wavefunctions.append(jnp.array(wavefunction))
            return jnp.array(wavefunctions)


"""
Cat Code Qubit
"""

from typing import Tuple

from jaxquantum.codes.base import BosonicQubit
import jaxquantum as jqt

from jax import config
import jax.numpy as jnp

config.update("jax_enable_x64", True)


class CatQubit(BosonicQubit):
    """
    Cat Qubit Class.
    """
    name = "cat"

    @property
    def _non_device_params(self):
        param_list = super()._non_device_params
        param_list.append("alpha")
        return param_list

    def _params_validation(self):
        super()._params_validation()
        if "alpha" not in self.params:
            self.params["alpha"] = 2

    def _get_basis_z(self) -> Tuple[jqt.Qarray, jqt.Qarray]:
        """
        Construct basis states |+-x>, |+-y>, |+-z>
        """
        N = self.params["N"]
        a = self.params["alpha"]
        plus_z = jqt.unit(jqt.coherent(N, a) + jqt.coherent(N, -1.0 * a))
        minus_z = jqt.unit(jqt.coherent(N, 1.0j * a) + jqt.coherent(N, -1.0j * a))
        return plus_z, minus_z



"""
Cat Code Qubit
"""

from typing import Tuple

from jaxquantum.codes.base import BosonicQubit
import jaxquantum as jqt

import jax.numpy as jnp
import jax.scipy as jsp


class GKPQubit(BosonicQubit):
    """
    GKP Qubit Class.
    """
    name = "gkp"

    def _params_validation(self):
        super()._params_validation()

        if "delta" not in self.params:
            self.params["delta"] = 0.25
        self.params["l"] = 2.0 * jnp.sqrt(jnp.pi)
        s_delta = jnp.sinh(self.params["delta"] ** 2)
        self.params["epsilon"] = s_delta * self.params["l"]

    def _gen_common_gates(self) -> None:
        """
        Overriding this method to add additional common gates.
        """
        super()._gen_common_gates()

        # phase space
        self.common_gates["x"] = (
            self.common_gates["a_dag"] + self.common_gates["a"]
        ) / jnp.sqrt(2.0)
        self.common_gates["p"] = (
            1.0j * (self.common_gates["a_dag"] - self.common_gates["a"]) / jnp.sqrt(2.0)
        )

        # finite energy
        self.common_gates["E"] = jqt.expm(
            -self.params["delta"] ** 2
            * self.common_gates["a_dag"]
            @ self.common_gates["a"]
        )
        self.common_gates["E_inv"] = jqt.expm(
            self.params["delta"] ** 2
            * self.common_gates["a_dag"]
            @ self.common_gates["a"]
        )

        # axis
        x_axis, z_axis = self._get_axis()
        y_axis = x_axis + z_axis

        # gates
        X_0 = jqt.expm(1.0j * self.params["l"] / 2.0 * z_axis)
        Z_0 = jqt.expm(1.0j * self.params["l"] / 2.0 * x_axis)
        Y_0 = 1.0j * X_0 @ Z_0
        self.common_gates["X_0"] = X_0
        self.common_gates["Z_0"] = Z_0
        self.common_gates["Y_0"] = Y_0
        self.common_gates["X"] = self._make_op_finite_energy(X_0)
        self.common_gates["Z"] = self._make_op_finite_energy(Z_0)
        self.common_gates["Y"] = self._make_op_finite_energy(Y_0)

        # symmetric stabilizers and gates
        self.common_gates["Z_s_0"] = self._symmetrized_expm(
            1.0j * self.params["l"] / 2.0 * x_axis
        )
        self.common_gates["S_x_0"] = self._symmetrized_expm(
            1.0j * self.params["l"] * z_axis
        )
        self.common_gates["S_z_0"] = self._symmetrized_expm(
            1.0j * self.params["l"] * x_axis
        )
        self.common_gates["S_y_0"] = self._symmetrized_expm(
            1.0j * self.params["l"] * y_axis
        )

    def _get_basis_z(self) -> Tuple[jqt.Qarray, jqt.Qarray]:
        """
        Construct basis states |+-x>, |+-y>, |+-z>.
        step 1: use ideal GKP stabilizers to find ideal GKP |+z> state
        step 2: make ideal eigenvector finite energy
            We want the groundstate of H = E H_0 E⁻¹.
            So, we can begin by find the groundstate of H_0 -> |λ₀⟩
            Then, we know that E|λ₀⟩ = |λ⟩ is the groundstate of H.
            pf. H|λ⟩ = (E H_0 E⁻¹)(E|λ₀⟩) = E H_0 |λ₀⟩ = λ₀ (E|λ₀⟩) = λ₀|λ⟩

        TODO (if necessary):
            Alternatively, we could construct a hamiltonian using
            finite energy stabilizers S_x, S_y, S_z, Z_s. However,
            this would make H = - S_x - S_y - S_z - Z_s non-hermitian.
            Currently, JAX does not support derivatives of jnp.linalg.eig,
            while it does support derivatives of jnp.linalg.eigh.
            Discussion: https://github.com/google/jax/issues/2748
        """

        # step 1: use ideal GKP stabilizers to find ideal GKP |+z> state
        H_0 = (
            -self.common_gates["S_x_0"]
            - self.common_gates["S_y_0"]
            - self.common_gates["S_z_0"]
            - self.common_gates["Z_s_0"]  # bosonic |+z> state
        )

        _, vecs = jnp.linalg.eigh(H_0.data)
        gstate_ideal = jqt.Qarray.create(vecs[:, 0])

        # step 2: make ideal eigenvector finite energy
        gstate = self.common_gates["E"] @ gstate_ideal

        N = self.params["N"]
        plus_z = jqt.unit(gstate)
        minus_z = jqt.unit(self.common_gates["X"] @ plus_z)
        return plus_z, minus_z

    # utils
    # ======================================================
    def _get_axis(self):
        x_axis = self.common_gates["x"]
        z_axis = -self.common_gates["p"]
        return x_axis, z_axis

    def _make_op_finite_energy(self, op):
        return self.common_gates["E"] @ op @ self.common_gates["E_inv"]

    def _symmetrized_expm(self, op):
        return (jqt.expm(op) + jqt.expm(-1.0 * op)) / 2.0

    # gates
    # ======================================================
    @property
    def x_U(self) -> jqt.Qarray:
        return self.common_gates["X"]

    @property
    def y_U(self) -> jqt.Qarray:
        return self.common_gates["Y"]

    @property
    def z_U(self) -> jqt.Qarray:
        return self.common_gates["Z"]


class RectangularGKPQubit(GKPQubit):
    def _params_validation(self):
        super()._params_validation()
        if "a" not in self.params:
            self.params["a"] = 0.8

    def _get_axis(self):
        a = self.params["a"]
        x_axis = a * self.common_gates["x"]
        z_axis = -1 / a * self.common_gates["p"]
        return x_axis, z_axis


class SquareGKPQubit(GKPQubit):
    def _params_validation(self):
        super()._params_validation()
        self.params["a"] = 1.0


class HexagonalGKPQubit(GKPQubit):
    def _get_axis(self):
        a = jnp.sqrt(2 / jnp.sqrt(3))
        x_axis = a * (
            jnp.sin(jnp.pi / 3.0) * self.common_gates["x"]
            + jnp.cos(jnp.pi / 3.0) * self.common_gates["p"]
        )
        z_axis = a * (-self.common_gates["p"])
        return x_axis, z_axis


## Citations

# Stabilization of Finite-Energy Gottesman-Kitaev-Preskill States
# Baptiste Royer, Shraddha Singh, and S. M. Girvin
# Phys. Rev. Lett. 125, 260509 – Published 31 December 2020

# Quantum error correction of a qubit encoded in grid states of an oscillator.
# Campagne-Ibarcq, P., Eickbusch, A., Touzard, S. et al.
# Nature 584, 368–372 (2020).



""" dims. """

from typing import List, Tuple
from copy import deepcopy
from math import prod
from jax import Array

from enum import Enum


DIMS_TYPE = List[List[int]]

def isket_dims(dims: DIMS_TYPE) -> bool:
    return prod(dims[1]) == 1

def isbra_dims(dims: DIMS_TYPE) -> bool:
    return prod(dims[0]) == 1

def isop_dims(dims: DIMS_TYPE) -> bool:
    return prod(dims[1]) == prod(dims[0])

def ket_from_op_dims(dims: DIMS_TYPE) -> DIMS_TYPE:
    return (dims[0], tuple([1 for _ in dims[1]]))


def check_dims(dims: Tuple[Tuple[int]], bdims: Tuple[int], data_shape: Array) -> bool:
    
    if len(data_shape) == 1 and data_shape[0] == 0:
        # E.g. empty list of operators
        assert bdims == (0,)
        assert dims == ((),())
        return 

    assert bdims == data_shape[:-2], "Data shape should be consistent with dimensions."
    assert data_shape[-2] == prod(dims[0]), "Data shape should be consistent with dimensions."
    assert data_shape[-1] == prod(dims[1]), "Data shape should be consistent with dimensions."

class Qdims:
    def __init__(self, dims):
        self._dims = deepcopy(dims)
        self._dims = (tuple(self._dims[0]), tuple(self._dims[1]))
        self._qtype = Qtypes.from_dims(self._dims)

    @property
    def dims(self):
        return self._dims

    @property
    def from_(self):
        return self._dims[1]
    
    @property
    def to_(self):
        return self._dims[0]
    
    @property
    def qtype(self):
        return self._qtype

    def __str__(self):
        return str(self.dims)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (self.dims == other.dims) and (self.qtype == other.qtype)

    def __ne__(self, other):
        return (self.dims != other.dims) or (self.qtype != other.qtype)

    def __hash__(self):
        return hash(self.dims)
    
    def __matmul__(self, other):
        if self.from_ != other.to_:
            raise TypeError(f"incompatible dimensions {self} and {other}")

        new_dims = [self.to_, other.from_]
        return Qdims(new_dims)


class Qtypes(str, Enum):
    ket = "ket"
    bra = "bra"
    oper = "oper"

    @classmethod
    def from_dims(cls, dims: Array):
        if isket_dims(dims):
            return cls.ket
        if isbra_dims(dims):
            return cls.bra
        if isop_dims(dims):
            return cls.oper
        raise ValueError("Invalid data shape")

    @classmethod
    def from_str(cls, string: str):
        return cls(string)

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return self.value != other.value

    def __hash__(self):
        return hash(self.value)


# /Users/phionx/Github/qc/EQuS/bosonic/jax/jaxquantum/test_example.py

import pytest

import sys
import os

# Add the jaxquantum directory to the sys.path
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import jaxquantum as jqt
import jax.numpy as jnp
import qutip as qt

minimum_version_for_tests = "0.2.0"


# Version
# ========================================
def test_conversions():
    a = jqt.basis(4, 3)
    assert jqt.jqt2qt(a) == qt.basis(4, 3)
    assert jqt.jqt2qt(None) is None

    a = jqt.displace(4, jnp.array([[1.0,2,3],[4,5,6]]))
    assert jqt.jqt2qt(a)[1][2] == jqt.jqt2qt(jqt.displace(4,6))

    b = jqt.displace(4,1.2)
    assert jqt.jnp2jqt(b.data) == b
    assert jqt.jnp2jqt(b.data, dims=(4,)) == b

    c = jqt.basis(4,1)
    assert jqt.jnp2jqt(c.data, dims=(4,)) == c

    assert jqt.qt2jqt(c) == c
    assert jqt.qt2jqt(jqt.jqt2qt(c)) == c


# ========================================

# Helpers
# ========================================

def test_overlap():
    a = jqt.basis(3,0)
    b = jqt.basis(3,1)
    n = jqt.num(3)

    assert jqt.overlap(a,b) == 0
    assert jqt.overlap(a,n) == 0
    assert jqt.overlap(n,b) == 1
    assert jqt.overlap(n,a.to_dm()) == 0
    assert jqt.overlap(b.to_dm(), n) == 1


# ========================================

# Utils
# ========================================
def test_device():
    a = {"test1": 1.0, "test2": jnp.array([1,2]), "test3": "wow"}
    jqt.device_put_params(a, non_device_params=["test3"])

def test_comb():
    assert jnp.abs(jqt.comb(5,2) - 10) < 1e-7

def test_iso_transforms():
    a = jqt.displace(3,1)
    assert jnp.all(jqt.real_to_complex_iso_matrix(jqt.complex_to_real_iso_matrix(a.data)) == a.data)
    
    b = jqt.coherent(3,1)
    assert jnp.all(jqt.real_to_complex_iso_vector(jqt.complex_to_real_iso_vector(a.data)) == a.data)

    a_iso = jqt.complex_to_real_iso_matrix(a.data)
    assert jnp.all(jqt.imag_times_iso_matrix(a_iso) == jqt.complex_to_real_iso_matrix((1j*a).data))
    assert jnp.all(jqt.conj_transpose_iso_matrix(a_iso) == jqt.complex_to_real_iso_matrix((a.dag()).data))

    b_iso = jqt.complex_to_real_iso_vector(b.data)
    assert jnp.all(jqt.imag_times_iso_vector(b_iso) == jqt.complex_to_real_iso_vector((1j*b).data))

# ========================================

# Operators

# ========================================

def test_misc_operators():
    assert jnp.max(jnp.abs(jqt.sigmay().data - jnp.array([[0.0, -1.0j], [1.0j, 0.0]]))) < 1e-12

    assert jnp.max(jnp.abs(jqt.hadamard().data - jnp.array([[1, 1], [1, -1]]) / jnp.sqrt(2)))  < 1e-12

    assert jnp.max(jnp.abs(jqt.sigmam().data - jnp.array([[0.0, 0.0], [1.0, 0.0]]))) < 1e-12

    assert jnp.max(jnp.abs(jqt.sigmap().data - jnp.array([[0.0, 1.0], [0.0, 0.0]]))) < 1e-12

    assert jnp.max(jnp.abs((jqt.qubit_rotation(0,0,0,0).data - jqt.identity(2).data))) < 1e-12

    assert jnp.max(jnp.abs(jqt.identity_like(jqt.sigmax()).data - jqt.identity(2).data)) < 1e-12

    state_val = jqt.thermal(3,0.1).data[2,2] 
    calc_val = jnp.exp(-0.1*2)/(jnp.exp(-0.1*0)+jnp.exp(-0.1*1)+jnp.exp(-0.1*2))
    assert jnp.abs(state_val-calc_val) < 1e-7

    assert jqt.thermal(10,jnp.inf) == jqt.basis(10,0).to_dm()

    assert jqt.basis_like((jqt.identity(2)^jqt.identity(3)), [1,0]) == (jqt.basis(2,1)^jqt.basis(3,0)) 
# ========================================


"""
Generic Bosonic Mode Class
"""

from typing import Tuple

from jaxquantum.codes.base import BosonicQubit
import jaxquantum as jqt

from jax import config
import jax.numpy as jnp

config.update("jax_enable_x64", True)


class BosonicMode(BosonicQubit):
    """
    FockQubit
    """

    def _params_validation(self):
        super()._params_validation()

    def _get_basis_z(self) -> Tuple[jqt.Qarray, jqt.Qarray]:
        """
        Construct basis states |+-x>, |+-y>, |+-z>
        """
        N = int(self.params["N"])
        plus_z = jqt.basis(N, 0)
        minus_z = jqt.basis(N, 1)
        return plus_z, minus_z



import pytest

import sys
import os

# Add the jaxquantum directory to the sys.path
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import jaxquantum as jqt
import jax.numpy as jnp
import jaxquantum.codes as jqtb

# GKP Code
# --

def test_gkp_code():
    gkp_qubit = jqtb.GKPQubit({"N":50})




""" ATS."""

from flax import struct
from jax import config

import jax.numpy as jnp

from jaxquantum.devices.superconducting.flux_base import FluxDevice
from jaxquantum.core.operators import identity, destroy, create
from jaxquantum.core.qarray import cosm, sinm

config.update("jax_enable_x64", True)

@struct.dataclass
class ATS(FluxDevice):        
    """
    ATS Device.
    """
    
    def common_ops(self):
        """ Written in the linear basis. """
        ops = {}
        
        N = self.N_pre_diag
        ops["id"] = identity(N)
        ops["a"] = destroy(N)
        ops["a_dag"] = create(N)
        ops["phi"] = self.phi_zpf()*(ops["a"] + ops["a_dag"])  
        ops["n"] = 1j * self.n_zpf() * (ops["a_dag"] - ops["a"])
        return ops

    def phi_zpf(self):
        """Return Phase ZPF."""
        return (2*self.params["Ec"]/self.params["El"])**(.25)

    def n_zpf(self):
        """Return Charge ZPF."""
        return (self.params["El"]/(32*self.params["Ec"]))**(.25)
    
    def get_linear_ω(self):
        """Get frequency of linear terms."""
        return jnp.sqrt(8*self.params["El"]*self.params["Ec"])
    
    def get_H_linear(self):
        """Return linear terms in H."""
        w = self.get_linear_ω()
        return w*(self.linear_ops["a_dag"]@self.linear_ops["a"] + 0.5 * self.linear_ops["id"])

    @staticmethod
    def get_H_nonlinear_static(phi_op, Ej, dEj, Ej2, phi_sum, phi_delta):
        cos_phi_op = cosm(phi_op)
        sin_phi_op = sinm(phi_op)

        cos_2phi_op = cos_phi_op @ cos_phi_op - sin_phi_op @ sin_phi_op
        sin_2phi_op = 2 * cos_phi_op @ sin_phi_op

        H_nl_Ej = - 2 * Ej * (
            cos_phi_op * jnp.cos(2 * jnp.pi * phi_delta) 
            - sin_phi_op * jnp.sin(2 * jnp.pi * phi_delta)
        ) * jnp.cos(2 * jnp.pi * phi_sum)
        H_nl_dEj = 2 * dEj * (
            sin_phi_op * jnp.cos(2 * jnp.pi * phi_delta)
            + cos_phi_op * jnp.sin(2 * jnp.pi * phi_delta)
        ) * jnp.sin(2 * jnp.pi * phi_sum) 
        H_nl_Ej2 = 2 * Ej2 * (
            cos_2phi_op * jnp.cos(2 * 2 * jnp.pi * phi_delta)
            - sin_2phi_op * jnp.sin(2 * 2 * jnp.pi * phi_delta)
        ) * jnp.cos(2 * 2 * jnp.pi * phi_sum)
        
        H_nl = H_nl_Ej + H_nl_dEj + H_nl_Ej2


        # id_op = jqt.identity_like(phi_op)
        # phi_delta_ext_op = self.params["phi_delta_ext"] * id_op
        # H_nl_old = - 2 * Ej * jqt.cosm(phi_op + 2 * jnp.pi * phi_delta_ext_op) * jnp.cos(2 * jnp.pi * self.params["phi_sum_ext"])
        # H_nl_old += 2 * dEj * jqt.sinm(phi_op + 2 * jnp.pi * phi_delta_ext_op) * jnp.sin(2 * jnp.pi * self.params["phi_sum_ext"]) 
        # H_nl_old += 2 * Ej2 * jqt.cosm(2*phi_op + 2 * 2 * jnp.pi * phi_delta_ext_op) * jnp.cos(2 * 2 * jnp.pi * self.params["phi_sum_ext"])
        
        return H_nl

    def get_H_nonlinear(self, phi_op):
        """Return nonlinear terms in H."""

        Ej = self.params["Ej"]
        dEj = self.params["dEj"]
        Ej2 = self.params["Ej2"]

        phi_sum = self.params["phi_sum_ext"]
        phi_delta = self.params["phi_delta_ext"]

        return ATS.get_H_nonlinear_static(
            phi_op, 
            Ej, 
            dEj, 
            Ej2, 
            phi_sum, 
            phi_delta
        )
        

    def get_H_full(self):
        """Return full H in linear basis."""
        id_op = self.linear_ops["id"]
        phi_b = self.linear_ops["phi"]
        H_nl = self.get_H_nonlinear(phi_b)
        H = self.get_H_linear() + H_nl
        return H

    def potential(self, phi):
        """Return potential energy for a given phi."""

        phi_delta_ext = self.params["phi_delta_ext"]
        phi_sum_ext = self.params["phi_sum_ext"]

        V = 0.5 * self.params["El"] * (2 * jnp.pi * phi) ** 2
        V += - 2 * self.params["Ej"] * jnp.cos(2 * jnp.pi * (phi + phi_delta_ext)) * jnp.cos(2 * jnp.pi * phi_sum_ext)
        V += 2 * self.params["dEj"] * jnp.sin(2 * jnp.pi * (phi + phi_delta_ext)) * jnp.sin(2 * jnp.pi * phi_sum_ext)
        V += 2 * self.params["Ej2"] * jnp.cos(2 * 2 * jnp.pi * (phi + phi_delta_ext)) * jnp.cos(2 * 2 * jnp.pi * phi_sum_ext)

        return V


""" States. """

from typing import List
from jax import config
from math import prod

import jax.numpy as jnp
from jax.nn import one_hot

from jaxquantum.core.qarray import Qarray, Qtypes, tensor

config.update("jax_enable_x64", True)



def sigmax() -> Qarray:
    """σx

    Returns:
        σx Pauli Operator
    """
    return Qarray.create(jnp.array([[0.0, 1.0], [1.0, 0.0]]))


def sigmay() -> Qarray:
    """σy

    Returns:
        σy Pauli Operator
    """
    return Qarray.create(jnp.array([[0.0, -1.0j], [1.0j, 0.0]]))


def sigmaz() -> Qarray:
    """σz

    Returns:
        σz Pauli Operator
    """
    return Qarray.create(jnp.array([[1.0, 0.0], [0.0, -1.0]]))


def hadamard() -> Qarray:
    """H

    Returns:
        H: Hadamard gate
    """
    return Qarray.create(jnp.array([[1, 1], [1, -1]]) / jnp.sqrt(2))

def sigmam() -> Qarray:
    """σ-

    Returns:
        σ- Pauli Operator
    """
    return Qarray.create(jnp.array([[0.0, 0.0], [1.0, 0.0]]))


def sigmap() -> Qarray:
    """σ+

    Returns:
        σ+ Pauli Operator
    """
    return Qarray.create(jnp.array([[0.0, 1.0], [0.0, 0.0]]))


def qubit_rotation(theta: float, nx, ny, nz) -> Qarray:
    """Single qubit rotation.

    Args:
        theta: rotation angle.
        nx: rotation axis x component.
        ny: rotation axis y component.
        nz: rotation axis z component.

    Returns:
        Single qubit rotation operator.
    """
    return jnp.cos(theta / 2) * identity(2) - 1j * jnp.sin(theta / 2) * (nx * sigmax() + ny * sigmay() + nz * sigmaz())

def destroy(N) -> Qarray:
    """annihilation operator

    Args:
        N: Hilbert space size

    Returns:
        annilation operator in Hilber Space of size N
    """
    return Qarray.create(jnp.diag(jnp.sqrt(jnp.arange(1, N)), k=1))


def create(N) -> Qarray:
    """creation operator

    Args:
        N: Hilbert space size

    Returns:
        creation operator in Hilber Space of size N
    """
    return Qarray.create(jnp.diag(jnp.sqrt(jnp.arange(1, N)), k=-1))


def num(N) -> Qarray:
    """Number operator

    Args:
        N: Hilbert Space size

    Returns:
        number operator in Hilber Space of size N
    """
    return Qarray.create(jnp.diag(jnp.arange(N)))


def identity(*args, **kwargs) -> Qarray:
    """Identity matrix.

    Returns:
        Identity matrix.
    """
    return Qarray.create(jnp.eye(*args, **kwargs))

def identity_like(A) -> Qarray:
    """Identity matrix with the same shape as A.

    Args:
        A: Matrix.

    Returns:
        Identity matrix with the same shape as A.
    """
    space_dims = A.space_dims 
    total_dim = prod(space_dims)
    return Qarray.create(jnp.eye(total_dim, total_dim), dims=[space_dims, space_dims])

def displace(N, α) -> Qarray:
    """Displacement operator

    Args:
        N: Hilbert Space Size
        α: Phase space displacement

    Returns:
        Displace operator D(α)
    """
    a = destroy(N)
    return (α * a.dag() - jnp.conj(α) * a).expm()


# States ---------------------------------------------------------------------

def basis(N: int, k: int):
    """Creates a |k> (i.e. fock state) ket in a specified Hilbert Space.

    Args:
        N: Hilbert space dimension
        k: fock number

    Returns:
        Fock State |k>
    """
    return Qarray.create(one_hot(k, N).reshape(N, 1))


def coherent(N: int, α: complex) -> Qarray:
    """Coherent state.

    Args:
        N: Hilbert Space Size.
        α: coherent state amplitude.

    Return:
        Coherent state |α⟩.
    """
    return displace(N, α) @ basis(N, 0)


def thermal(N: int, beta: float) -> Qarray:
    """Thermal state.

    Args:
        N: Hilbert Space Size.
        beta: thermal state inverse temperature.

    Return:
        Thermal state.
    """

    return Qarray.create(jnp.where(jnp.isposinf(beta),
                     basis(N, 0).to_dm().data,
                     jnp.diag(jnp.exp(-beta*jnp.linspace(0, N-1,N))))).unit()


def basis_like(A: Qarray, ks: List[int]) -> Qarray:
    """Creates a |k> (i.e. fock state) ket with the same space dims as A.

    Args:
        A: state or operator.
        k: fock number.

    Returns:
        Fock State |k> with the same space dims as A.
    """
    space_dims = A.space_dims 
    assert len(space_dims) == len(ks), "len(ks) must be equal to len(space_dims)"
    
    kets = []
    for j, k in enumerate(ks):
        kets.append(basis(space_dims[j], k))
    return tensor(*kets)


""" System."""

from functools import partial
from typing import List, Optional, Dict, Any, Union
import math

from flax import struct
from jax import vmap, Array
from jax import config

import jax.numpy as jnp

from jaxquantum.devices.base.base import Device
from jaxquantum.devices.superconducting.drive import Drive
from jaxquantum.core.qarray import Qarray, tensor
from jaxquantum.core.operators import identity

config.update("jax_enable_x64", True)


def calculate_eig(Ns, H: Qarray):
    N_tot = math.prod(Ns)
    edxs = jnp.arange(N_tot)

    vals, kets = jnp.linalg.eigh(H.data)
    kets = kets.T

    def calc_quantum_number(edx):
        argmax = jnp.argmax(jnp.abs(kets[edx]))
        val = vals[edx]  # - vals[0]
        return val, argmax, kets[edx]

    quantum_numbers = vmap(calc_quantum_number)(edxs)

    def calc_order(edx):
        indx = jnp.argmin(jnp.abs(edx - quantum_numbers[1]))
        return quantum_numbers[0][indx], quantum_numbers[2][indx]

    Es, kets = vmap(calc_order)(edxs)

    kets = jnp.reshape(kets, (N_tot, N_tot, 1))
    kets = Qarray.create(kets)
    kets = kets.reshape_qdims(*Ns)
    kets = kets.reshape_bdims(*Ns)

    return (
        jnp.reshape(Es, Ns),
        kets,
    )


def promote(op: Qarray, device_num, Ns):
    I_ops = [identity(N) for N in Ns]
    return tensor(*I_ops[:device_num], op, *I_ops[device_num + 1 :])


@struct.dataclass
class System:
    Ns: List[int] = struct.field(pytree_node=False)
    devices: List[Union[Device, Drive]]
    couplings: List[Array]
    params: Dict[str, Any]

    @classmethod
    def create(
        cls,
        devices: List[Union[Device, Drive]],
        couplings: Optional[List[Array]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        labels = [device.label for device in devices]
        unique_labels = set(labels)
        if len(labels) != len(unique_labels):
            raise ValueError("Devices must have unique labels.")
            
        Ns = tuple([device.N for device in devices])
        couplings = couplings if couplings is not None else []
        params = params if params is not None else {}
        return cls(Ns, devices, couplings, params)

    def promote(self, op, device_num):
        return promote(op, device_num, self.Ns)

    def get_H_bare(self):
        I_ops = [identity(N) for N in self.Ns]
        H = 0
        for j, device in enumerate(self.devices):
            H += self.promote(device.get_H(), j)
        return H

    def get_H_couplings(self):
        H = 0
        for coupling in self.couplings:
            H += coupling
        return H
    
    def get_H(self):
        H_bare = self.get_H_bare()
        H_couplings = self.get_H_couplings()
        return H_bare + H_couplings

    def calculate_eig(self):
        H = self.get_H()
        return calculate_eig(self.Ns, H)



""" Tunable Transmon."""

from flax import struct
from jax import config
import jax.numpy as jnp

from jaxquantum.devices.superconducting.transmon import Transmon

config.update("jax_enable_x64", True)

@struct.dataclass
class TunableTransmon(Transmon):
    """
    Tunable Transmon Device.
    """
    
    @property
    def Ej(self):
        Ejsum = (self.params["Ej1"] + self.params["Ej2"])
        phi_ext = 2 * jnp.pi * self.params["phi_ext"]
        gamma = self.params["Ej2"]/self.params["Ej1"]
        d = (gamma - 1)/(gamma + 1)
        external_flux_factor = jnp.abs(jnp.sqrt(jnp.cos(phi_ext/2)**2 + d**2 * jnp.sin(phi_ext/2)**2))
        return Ejsum * external_flux_factor


"""
Setup
"""
import os

from setuptools import setup, find_namespace_packages

DIST_NAME = "jaxquantum"
PACKAGE_NAME = "jaxquantum"

REQUIREMENTS = [
    "numpy",
    "matplotlib",
    "qutip",
    "jax[cpu]",
    "diffrax",
    "flax",
    "tqdm",
    "pytest"
]

EXTRA_REQUIREMENTS = {
    "dev": [
        "jupyterlab>=3.1.0",
        "mypy",
        "pylint",
        "black",
        "coverage",
        "pytest"
    ],
    "docs": [
        "mkdocs",
        "mkdocs-material",
        "mkdocs-literate-nav",
        "mkdocs-section-index",
        "mkdocs-gen-files",
        "mkdocstrings-python",
        "mkdocs-jupyter"
    ]
}

# Read long description from README.
README_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md")
with open(README_PATH) as readme_file:
    README = readme_file.read()

version_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), PACKAGE_NAME, "VERSION.txt")
)

with open(version_path, "r") as fd:
    version_str = fd.read().rstrip()

setup(
    name=DIST_NAME,
    version=version_str,
    description=DIST_NAME,
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/EQuS/jaxquantum",
    author="Shantanu Jha, Shoumik Chowdhury, Max Hays",
    author_email="shantanu.rajesh.jha@gmail.com",
    license="Apache-2.0",
    packages=find_namespace_packages(exclude=["experiments*"]),
    install_requires=REQUIREMENTS,
    extras_require=EXTRA_REQUIREMENTS,
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering",
    ],
    keywords="quantum computing",
    python_requires=">=3.7",
    project_urls={
        "Documentation": "https://equs.github.io/jaxquantum",
        "Source Code": "https://github.com/EQuS/jaxquantum",
        "Tutorials": "https://github.com/EQuS/jaxquantum/tree/main/tutorials",
    },
    include_package_data=True,
)



""" QArray. """

from __future__ import annotations

import functools
from flax import struct
from enum import Enum
from jax import Array, config, vmap
from typing import List, Optional, Union


from math import prod
from copy import deepcopy
from numbers import Number
from numpy import ndarray
import jax.numpy as jnp
import jax.scipy as jsp

from jaxquantum.core.settings import SETTINGS
from jaxquantum.utils.utils import robust_isscalar
from jaxquantum.core.dims import Qtypes, Qdims, check_dims, isket_dims, isbra_dims, isop_dims, ket_from_op_dims

config.update("jax_enable_x64", True)

def tidy_up(data, atol):
    data_re = jnp.real(data)
    data_im = jnp.imag(data)
    data_re_mask = jnp.abs(data_re) > atol
    data_im_mask = jnp.abs(data_im) > atol
    data_new = data_re * data_re_mask + 1j * data_im * data_im_mask
    return data_new


@struct.dataclass # this allows us to send in and return Qarray from jitted functions
class Qarray:
    _data: Array
    _qdims: Qdims = struct.field(pytree_node=False)
    _bdims: Tuple[int] = struct.field(pytree_node=False)


    # Initialization ----
    @classmethod
    def create(cls, data, dims=None, bdims=None):

        # Step 1: Prepare data ----
        data = jnp.asarray(data)

        if len(data.shape) == 1 and data.shape[0] > 0:
            data = data.reshape(data.shape[0], 1)

        if len(data.shape) >= 2:
            if data.shape[-2] != data.shape[-1] and not (data.shape[-2] == 1 or data.shape[-1] == 1):
                data = data.reshape(*data.shape[:-1], data.shape[-1], 1)

        if bdims is not None:
            if len(data.shape) - len(bdims) == 1:
                data = data.reshape(*data.shape[:-1], data.shape[-1], 1)
        # ----

        # Step 2: Prepare dimensions ----
        if bdims is None:
            bdims = tuple(data.shape[:-2])

        if dims is None:
            dims = ((data.shape[-2],), (data.shape[-1],))

        dims = (tuple(dims[0]), tuple(dims[1]))
        
        check_dims(dims, bdims, data.shape)

        qdims = Qdims(dims)

        # NOTE: Constantly tidying up on Qarray creation might be a bit overkill.
        # It increases the compilation time, but only very slightly 
        # increased the runtime of the jit compiled function.
        # We could instead use this tidy_up where we think we need it.
        data = tidy_up(data, SETTINGS["auto_tidyup_atol"])

        return cls(data, qdims, bdims)

    # ----

    @classmethod
    def from_list(cls, qarr_list: List[Qarray]) -> Qarray:
        """ Create a Qarray from a list of Qarrays. """
        
        data = jnp.array([qarr.data for qarr in qarr_list])
        

        if len(qarr_list) == 0:
            dims = ((),())
            bdims = ()
        else:
            dims = qarr_list[0].dims
            bdims = qarr_list[0].bdims
        
        if not all(qarr.dims == dims and qarr.bdims == bdims for qarr in qarr_list):
            raise ValueError("All Qarrays in the list must have the same dimensions.")
            
        bdims = (len(qarr_list),) + bdims

        return cls.create(data, dims=dims, bdims=bdims)

    @classmethod
    def from_array(cls, qarr_arr) -> Qarray:
        """ Create a Qarray from a nested list of Qarrays. 
        
        Args:
            qarr_arr (list): nested list of Qarrays
        
        Returns:
            Qarray: Qarray object
        """
        if isinstance(qarr_arr, Qarray):
            return qarr_arr

        bdims = ()
        lvl = qarr_arr
        while not isinstance(lvl, Qarray):
            bdims = bdims + (len(lvl),)
            if len(lvl) > 0:
                lvl = lvl[0]
            else:
                break
        
        depth = len(bdims)

        def flat(lis):
            flatList = []
            for element in lis:
                if type(element) is list:
                    flatList += flat(element)
                else:
                    flatList.append(element)
            return flatList

        qarr_list = flat(qarr_arr)
        qarr = cls.from_list(qarr_list)
        qarr = qarr.reshape_bdims(*bdims)
        return qarr


    # Properties ----
    @property
    def qtype(self):
        return self._qdims.qtype
    
    @property
    def dtype(self):
        return self._data.dtype

    @property
    def dims(self):
        return self._qdims.dims

    @property
    def bdims(self):
        return self._bdims

    @property
    def qdims(self):
        return self._qdims
        
    @property
    def space_dims(self):
        if self.qtype in [Qtypes.oper, Qtypes.ket]:
            return self.dims[0]
        elif self.qtype == Qtypes.bra:
            return self.dims[1]
        else:
            # TODO: not reached for some reason
            raise ValueError("Unsupported qtype.")
        
    @property
    def data(self):
        return self._data
    
    @property
    def shaped_data(self):
        return self._data.reshape(self.bdims + self.dims[0] + self.dims[1])

    @property 
    def shape(self):
        return self.data.shape

    @property
    def is_batched(self):
        return len(self.bdims) > 0

    def __getitem__(self, index):
        if len(self.bdims) > 0:
            return Qarray.create(
                self.data[index],
                dims=self.dims,
            )
        else:
            raise ValueError("Cannot index a non-batched Qarray.")

    def reshape_bdims(self, *args):
        """ Reshape the batch dimensions of the Qarray. """
        new_bdims = tuple(args)

        if prod(new_bdims) == 0:
            new_shape = new_bdims 
        else:
            new_shape = new_bdims + (prod(self.dims[0]),) + (-1,)
        return Qarray.create(
            self.data.reshape(new_shape),
            dims=self.dims,
            bdims=new_bdims,
        )

    def space_to_qdims(self, space_dims: List[int]):
        
        if isinstance(space_dims[0], (list, tuple)):
            return space_dims

        if self.qtype in [Qtypes.oper, Qtypes.ket]:
            return (tuple(space_dims), tuple([1 for _ in range(len(space_dims))]))
        elif self.qtype == Qtypes.bra:
            return (tuple([1 for _ in range(len(space_dims))]), tuple(space_dims))
        else:
            raise ValueError("Unsupported qtype for space_to_qdims conversion.")

    def reshape_qdims(self, *args):
        """ Reshape the quantum dimensions of the Qarray. 
        
        Note that this does not take in qdims but rather the new Hilbert space dimensions.

        Args:
            *args: new Hilbert dimensions for the Qarray.
        
        Returns:
            Qarray: reshaped Qarray.
        """

        new_space_dims = tuple(args)
        current_space_dims = self.space_dims
        assert prod(new_space_dims) == prod(current_space_dims)


        new_qdims = self.space_to_qdims(new_space_dims)
        new_bdims = self.bdims

        return Qarray.create(
            self.data,
            dims=new_qdims,
            bdims=new_bdims
        )
        
    def resize(self, new_shape):
        """ Resize the Qarray to a new shape. 
        
        TODO: review and maybe deprecate this method.
        """
        dims = self.dims
        data = jnp.resize(self.data, new_shape)
        return Qarray.create(
            data,
            dims=dims,
        )

    def __len__(self):
        """ Length of the Qarray. """
        if len(self.bdims) > 0:
            return self.data.shape[0]
        else:
            raise ValueError("Cannot get length of a non-batched Qarray.")


    def __eq__(self, other):
        if not isinstance(other, Qarray):
            raise ValueError("Cannot calculate equality of a Qarray with a non-Qarray.")

        if self.dims != other.dims:
            return False

        if self.bdims != other.bdims:
            return False

        return jnp.all(self.data == other.data)
    
    def __ne__(self, other):
        return not self.__eq__(other)
    # ----


    # Elementary Math ----
    def __matmul__(self, other):
        if not isinstance(other, Qarray):
            return NotImplemented
        _qdims_new = self._qdims @ other._qdims
        return Qarray.create(
            self.data @ other.data,
            dims=_qdims_new.dims,
        )
    
    # NOTE: not possible to reach this.
    # def __rmatmul__(self, other):        
    #     if not isinstance(other, Qarray):
    #         return NotImplemented

    #     _qdims_new = other._qdims @ self._qdims
    #     return Qarray.create(
    #         other.data @ self.data,
    #         dims=_qdims_new.dims,
    #     )
        
    
    def __mul__(self, other):
        if isinstance(other, Qarray):
            return self.__matmul__(other)
     

        other = other + 0.0j
        if not robust_isscalar(other) and len(other.shape) > 0: # not a scalar
            other = other.reshape(other.shape + (1,1))
            
        return Qarray.create(
            other * self.data,
            dims=self._qdims.dims,
        )

    def __rmul__(self, other):
        
        # NOTE: not possible to reach this.
        # if isinstance(other, Qarray):
        #     return self.__rmatmul__(other)
        
        return self.__mul__(other)

    def __neg__(self):
        return self.__mul__(-1)
    
    def __truediv__(self, other):
        """ For Qarray's, this only really makes sense in the context of division by a scalar. """

        if isinstance(other, Qarray):
            raise ValueError("Cannot divide a Qarray by another Qarray.")

        return self.__mul__(1/other)
    
    def __add__(self, other):
        if isinstance(other, Qarray):
            if self.dims != other.dims:
                msg = (
                    "Dimensions are incompatible: "
                    + repr(self.dims) + " and " + repr(other.dims)
                )
                raise ValueError(msg)
            return Qarray.create(self.data + other.data, dims=self.dims)

        if robust_isscalar(other) and other == 0:
            return self.copy()

        if self.data.shape[-2] == self.data.shape[-1]:
            other = other + 0.0j
            if not robust_isscalar(other) and len(other.shape) > 0: # not a scalar
                other = other.reshape(other.shape + (1,1))
            other = Qarray.create(other * jnp.eye(self.data.shape[-2], dtype=self.data.dtype), dims=self.dims)
            return self.__add__(other)
        
        return NotImplemented
    
    def __radd__(self, other):
        return self.__add__(other)
    
    def __sub__(self, other):    
        if isinstance(other, Qarray):
            if self.dims != other.dims:
                msg = (
                    "Dimensions are incompatible: "
                    + repr(self.dims) + " and " + repr(other.dims)
                )
                raise ValueError(msg)
            return Qarray.create(self.data - other.data, dims=self.dims)

        if robust_isscalar(other) and other == 0:
            return self.copy()

        if self.data.shape[-2] == self.data.shape[-1]:
            other = other + 0.0j
            if not robust_isscalar(other) and len(other.shape) > 0: # not a scalar
                other = other.reshape(other.shape + (1,1))
            other = Qarray.create(other * jnp.eye(self.data.shape[-2], dtype=self.data.dtype), dims=self.dims)
            return self.__sub__(other)
        
        return NotImplemented
        
    def __rsub__(self, other):
        return self.__neg__().__add__(other)
    
    def __xor__(self, other):
        if not isinstance(other, Qarray):
            return NotImplemented
        return tensor(self, other)

    def __rxor__(self, other):
        if not isinstance(other, Qarray):
            return NotImplemented
        return tensor(other, self)
    
    def __pow__(self, other):
        if not isinstance(other, int):
            return NotImplemented
        
        return powm(self, other)
    
    # ----

    # String Representation ----
    def _str_header(self):
        out = ", ".join([
            "Quantum array: dims = " + str(self.dims),
            "bdims = " + str(self.bdims),
            "shape = " + str(self._data.shape),
            "type = " + str(self.qtype),
        ])
        return out

    def __str__(self):
        return self._str_header() + "\nQarray data =\n" + str(self.data)
        
    @property
    def header(self):
        """ Print the header of the Qarray. """
        return self._str_header()

    def __repr__(self):
        return self.__str__()
    
    # ----

    # Utilities ----
    def copy(self, memo=None):
        # return Qarray.create(deepcopy(self.data), dims=self.dims)
        return self.__deepcopy__(memo)
    
    def __deepcopy__(self, memo):
        """ Need to override this when defininig __getattr__. """

        return Qarray(
            _data = deepcopy(self._data, memo=memo),
            _qdims = deepcopy(self._qdims, memo=memo),
            _bdims = deepcopy(self._bdims, memo=memo)
        )

    def __getattr__(self, method_name):

        if "__" == method_name[:2]:
            # NOTE: we return NotImplemented for binary special methods logic in python, plus things like __jax_array__
            return lambda *args, **kwargs: NotImplemented

        modules = [jnp, jnp.linalg, jsp, jsp.linalg]

        method_f = None 
        for mod in modules:
            method_f = getattr(mod, method_name, None)
            if method_f is not None:
                break

        if method_f is None:
            raise NotImplementedError(f"Method {method_name} does not exist. No backup method found in {modules}.")

        def func(*args, **kwargs): 
            res = method_f(self.data, *args, **kwargs)
            
            if getattr(res, "shape", None) is None or res.shape != self.data.shape:
                return res
            else:
                return Qarray.create(
                    res,
                    dims=self._qdims.dims
                )
        return func

    # ----

    # Conversions / Reshaping ----
    def dag(self):
        return dag(self)
    
    def to_dm(self):
        return ket2dm(self)
    
    def is_dm(self):
        return self.qtype == Qtypes.oper

    def is_vec(self):
        return self.qtype == Qtypes.ket or self.qtype == Qtypes.bra
    
    def to_ket(self):
        return to_ket(self)
    
    def transpose(self, *args):
        return transpose(self, *args)

    def keep_only_diag_elements(self):
        return keep_only_diag_elements(self)
    
    # ----

    # Math Functions ----
    def unit(self):
        return unit(self)

    def norm(self):
        return norm(self)

    def expm(self):
        return expm(self)

    def powm(self, n):
        return powm(self, n)
    
    def cosm(self):
        return cosm(self)
    
    def sinm(self):
        return sinm(self)
    
    def tr(self, **kwargs):
        return tr(self, **kwargs)

    def trace(self, **kwargs):
        return tr(self, **kwargs)

    def ptrace(self, indx):
        return ptrace(self, indx)
        
    def eigenstates(self):
        return eigenstates(self)

    def eigenenergies(self):
        return eigenenergies(self)
    
    def collapse(self, mode="sum"):
        return collapse(self, mode=mode)    
    # ----

ARRAY_TYPES = (Array, ndarray, Qarray)

# Qarray operations ---------------------------------------------------------------------

def collapse(qarr: Qarray, mode="sum") -> Qarray:
    """Collapse the Qarray.

    Args:
        qarr (Qarray): quantum array array
    
    Returns:
        Collapsed quantum array
    """
    if mode == "sum":
        if len(qarr.bdims) == 0:
            return self

        batch_axes = list(range(len(qarr.bdims)))
        return Qarray.create(
            jnp.sum(qarr.data, axis=batch_axes),
            dims=qarr.dims
        )
    

def transpose(qarr: Qarray, indices: List[int]) -> Qarray:
    """ Transpose the quantum array.

    Args:
        qarr (Qarray): quantum array
        *args: axes to transpose
    
    Returns:
        tranposed Qarray
    """

    indices = list(indices)

    shaped_data = qarr.shaped_data
    dims = qarr.dims
    bdims_indxs = list(range(len(qarr.bdims)))

    reshape_indices = indices + [j + len(dims[0]) for j in indices]

    reshape_indices = bdims_indxs + [j + len(bdims_indxs) for j in reshape_indices]

    shaped_data = shaped_data.transpose(reshape_indices)
    new_dims = (tuple([dims[0][j] for j in indices]), tuple([dims[1][j] for j in indices]))

    full_dims = prod(dims[0])
    full_data = shaped_data.reshape(*qarr.bdims, full_dims, -1)
    return Qarray.create(full_data, dims = new_dims)

def unit(qarr: Qarray) -> Qarray:
    """Normalize the quantum array.

    Args:
        qarr (Qarray): quantum array

    Returns:
        Normalized quantum array
    """
    data = qarr.data
    data = data / qarr.norm()
    return Qarray.create(data, dims=qarr.dims)

def norm(qarr: Qarray) -> float:
    data = qarr.data
    data_dag = qarr.dag().data

    if qarr.qtype == Qtypes.oper:
        evals, _ = jnp.linalg.eigh(data @ data_dag)
        rho_norm = jnp.sum(jnp.sqrt(jnp.abs(evals)))
        return rho_norm
    elif qarr.qtype in [Qtypes.ket, Qtypes.bra]:
        return jnp.linalg.norm(data)

def tensor(*args, **kwargs) -> Qarray:
    """Tensor product.

    Args:
        *args (Qarray): tensors to take the product of
        parallel (bool): if True, use parallel einsum for tensor product 
            true: [A,B] ^ [C,D] = [A^C, B^D]
            false: [A,B] ^ [C,D] = [A^C, A^D, B^C, B^D]

    Returns:
        Tensor product of given tensors

    """

    parallel = kwargs.pop("parallel", False)
    
    data = args[0].data
    dims = deepcopy(args[0].dims)
    dims_0 = dims[0]
    dims_1 = dims[1]
    for arg in args[1:]:

        if parallel:
            a = data 
            b = arg.data

            if len(a.shape) > len(b.shape):
                batch_dim = a.shape[:-2]
            elif len(a.shape) == len(b.shape):
                if prod(a.shape[:-2]) > prod(b.shape[:-2]):
                    batch_dim = a.shape[:-2]
                else:
                    batch_dim = b.shape[:-2]
            else:
                batch_dim = b.shape[:-2]

            data = jnp.einsum("...ij,...kl->...ikjl", a, b).reshape(*batch_dim, a.shape[-2] * b.shape[-2], -1)
        else:
            data = jnp.kron(data, arg.data, **kwargs)

        dims_0 = dims_0 + arg.dims[0]
        dims_1 = dims_1 + arg.dims[1]

    return Qarray.create(data, dims=(dims_0, dims_1))

def tr(qarr: Qarray, **kwargs) -> Array:
    """Full trace.

    Args:
        qarr (Qarray): quantum array    

    Returns:
        Full trace.
    """
    axis1 = kwargs.get("axis1", -2)
    axis2 = kwargs.get("axis2", -1)
    return jnp.trace(qarr.data, axis1=axis1, axis2=axis2, **kwargs)

def trace(qarr: Qarray, **kwargs) -> Array:
    """Full trace.

    Args:
        qarr (Qarray): quantum array    

    Returns:
        Full trace.
    """
    return tr(qarr, **kwargs)
    


def expm_data(data: Array, **kwargs) -> Array:
    """Matrix exponential wrapper.

    Returns:
        matrix exponential
    """
    return jsp.linalg.expm(data, **kwargs)

def expm(qarr: Qarray, **kwargs) -> Qarray:
    """Matrix exponential wrapper.

    Returns:
        matrix exponential
    """
    dims = qarr.dims
    data = expm_data(qarr.data, **kwargs)
    return Qarray.create(data, dims=dims)

def powm(qarr: Qarray, n: Union[int, float]) -> Qarray:
    """Matrix power.

    Args:
        qarr (Qarray): quantum array
        n (int): power

    Returns:
        matrix power
    """
    if type(n) == int:
        data_res = jnp.linalg.matrix_power(qarr.data, n)
    else:
        evalues, evectors = jnp.linalg.eig(qarr.data)
        if not (evalues >= 0).all():
            raise ValueError("Non-integer power of a matrix can only be "
                             "computed if the matrix is positive semi-definite."
                             "Got a matrix with a negative eigenvalue.")
        data_res = evectors * jnp.pow(evalues, n) @ jnp.linalg.inv(evectors)
    return Qarray.create(data_res, dims=qarr.dims)
    
def cosm_data(data: Array, **kwargs) -> Array:
    """Matrix cosine wrapper.

    Returns:
        matrix cosine
    """
    return (expm_data(1j*data) + expm_data(-1j*data))/2 

def cosm(qarr: Qarray) -> Qarray:
    """Matrix cosine wrapper.

    Args:
        qarr (Qarray): quantum array

    Returns:
        matrix cosine
    """
    dims = qarr.dims
    data = cosm_data(qarr.data)
    return Qarray.create(data, dims=dims)

def sinm_data(data: Array, **kwargs) -> Array:
    """Matrix sine wrapper.

    Args:
        data: matrix

    Returns:
        matrix sine
    """
    return (expm_data(1j*data) - expm_data(-1j*data))/(2j)

def sinm(qarr: Qarray) -> Qarray:
    dims = qarr.dims
    data = sinm_data(qarr.data)
    return Qarray.create(data, dims=dims)

def keep_only_diag_elements(qarr: Qarray) -> Qarray:
    if len(qarr.bdims) > 0:
        raise ValueError("Cannot keep only diagonal elements of a batched Qarray.")

    dims = qarr.dims
    data = jnp.diag(jnp.diag(qarr.data))
    return Qarray.create(data, dims=dims)

def to_ket(qarr: Qarray) -> Qarray:
    if qarr.qtype == Qtypes.ket:
        return qarr
    elif qarr.qtype == Qtypes.bra:
        return qarr.dag()
    else:
        raise ValueError("Can only get ket from a ket or bra.")
    
def eigenstates(qarr: Qarray) -> Qarray:
    """Eigenstates of a quantum array.

    Args:
        qarr (Qarray): quantum array

    Returns:
        eigenvalues and eigenstates
    """

    evals, evecs = jnp.linalg.eigh(qarr.data)
    idxs_sorted = jnp.argsort(evals, axis=-1)

    dims = ket_from_op_dims(qarr.dims)

    evals = jnp.take_along_axis(evals, idxs_sorted, axis=-1)
    evecs = jnp.take_along_axis(evecs, idxs_sorted[...,None,:], axis=-1)

    evecs = Qarray.create(
        evecs,
        dims=dims,
        bdims=evecs.shape[:-1],
    )

    return evals, evecs

def eigenenergies(qarr: Qarray) -> Array:
    """Eigenvalues of a quantum array.

    Args:
        qarr (Qarray): quantum array

    Returns:
        eigenvalues
    """

    evals = jnp.linalg.eigvalsh(qarr.data)
    return evals

# More quantum specific -----------------------------------------------------

def ptrace(qarr: Qarray, indx) -> Qarray:
    """Partial Trace.

    Args:
        rho: density matrix
        indx: index of quantum object to keep, rest will be partial traced out

    Returns:
        partial traced out density matrix

    TODO: Fix weird tracing errors that arise with reshape
    """

    qarr = ket2dm(qarr)
    rho = qarr.shaped_data
    dims = qarr.dims

    Nq = len(dims[0])

    indxs = [indx, indx + Nq]
    for j in range(Nq):
        if j == indx:
            continue
        indxs.append(j)
        indxs.append(j + Nq)

    bdims = qarr.bdims
    len_bdims = len(bdims)
    bdims_indxs = list(range(len_bdims))
    indxs = bdims_indxs + [j + len_bdims for j in indxs]
    rho = rho.transpose(indxs)

    for j in range(Nq - 1):
        rho = jnp.trace(rho, axis1=2+len_bdims, axis2=3+len_bdims)

    return Qarray.create(rho)

def dag(qarr: Qarray) -> Qarray:
    """Conjugate transpose.

    Args:
        qarr (Qarray): quantum array

    Returns:
        conjugate transpose of qarr
    """
    dims = qarr.dims[::-1]

    data = dag_data(qarr.data)
    
    return Qarray.create(data, dims=dims)

def dag_data(arr: Array) -> Array:
    """Conjugate transpose.

    Args:
        arr: operator

    Returns:
        conjugate of op, and transposes last two axes
    """
    # TODO: revisit this case...
    if len(arr.shape) == 1:
        return jnp.conj(arr)
        
    return jnp.moveaxis(
        jnp.conj(arr), -1, -2
    )  # transposes last two axes, good for batching

def ket2dm(qarr: Qarray) -> Qarray:
    """Turns ket into density matrix.
    Does nothing if already operator.

    Args:
        qarr (Qarray): qarr

    Returns:
        Density matrix
    """

    if qarr.qtype == Qtypes.oper:
        return qarr
    
    if qarr.qtype == Qtypes.bra:
        qarr = qarr.dag()

    return qarr @ qarr.dag()


# Data level operations ----

def is_dm_data(data: Array) -> bool:
    """Check if data is a density matrix.

    Args:
        data: matrix
    Returns:
        True if data is a density matrix
    """
    return data.shape[-2] == data.shape[-1]

def powm_data(data: Array, n: int) -> Array:
    """Matrix power.

    Args:
        data: matrix
        n: power

    Returns:
        matrix power
    """
    return jnp.linalg.matrix_power(data, n)


"""
Bosonic Code Qubits
"""

from .base import BosonicQubit
from .mode import BosonicMode
from .qubit import Qubit
from .cat import CatQubit
from .binomial import BinomialQubit
from .gkp import GKPQubit, HexagonalGKPQubit, SquareGKPQubit, RectangularGKPQubit



"""Solvers"""


from diffrax import diffeqsolve, ODETerm, SaveAt, PIDController, TqdmProgressMeter, NoProgressMeter
from functools import partial
from flax import struct
from jax import vmap, Array
from typing import Callable, List, Optional, Dict, Union
import diffrax
import jax.numpy as jnp
import jax.scipy as jsp
import warnings
import tqdm
import logging



from jaxquantum.core.qarray import Qarray, Qtypes, is_dm_data, dag_data
from jaxquantum.core.conversions import jnp2jqt
from jaxquantum.utils.utils import robust_isscalar
from jaxquantum.core.operators import identity_like
from jaxquantum.core.measurements import overlap

# ----

@struct.dataclass
class SolverOptions:
    progress_meter: bool = struct.field(pytree_node=False)
    solver: str = struct.field(pytree_node=False),
    max_steps: int = struct.field(pytree_node=False),
    

    @classmethod
    def create(cls, progress_meter: bool = True, solver: str = "Tsit5", max_steps: int = 100_000):
        return cls(progress_meter, solver, max_steps)


class CustomProgressMeter(TqdmProgressMeter):
    @staticmethod
    def _init_bar() -> tqdm.tqdm:
        bar_format = "{desc}: {percentage:3.0f}% |{bar}| [{elapsed}<{remaining}, {rate_fmt}{postfix}]"
        return tqdm.tqdm(total=100, bar_format=bar_format, unit='%', colour="MAGENTA", ascii="░▒█")
    
def solve(f, ρ0, tlist, args, solver_options: Optional[SolverOptions] = None):
    """ Gets teh desired solver from diffrax.

    Args:
        solver_options: dictionary with solver options

    Returns:
        solution 
    """

    # f and ts
    term = ODETerm(f)
    saveat = SaveAt(ts=tlist)

    # solver 
    solver_options = solver_options or SolverOptions.create()
    
    solver_name = solver_options.solver
    solver = getattr(diffrax, solver_name)()
    stepsize_controller = PIDController(rtol=1e-6, atol=1e-6)

    # solve!
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', UserWarning) # NOTE: suppresses complex dtype warning in diffrax
        sol = diffeqsolve(
            term,
            solver,
            t0=tlist[0],
            t1=tlist[-1],
            dt0=tlist[1] - tlist[0],
            y0=ρ0,
            saveat=saveat,
            stepsize_controller=stepsize_controller,
            args=args,
            max_steps=solver_options.max_steps,
            progress_meter=CustomProgressMeter() if solver_options.progress_meter else NoProgressMeter(),
        )   

    return sol

def mesolve(
    H: Union[Qarray, Callable[[float], Qarray]],
    rho0: Qarray,
    tlist: Array,
    c_ops: Optional[Qarray] = None,
    solver_options: Optional[SolverOptions] = None
) -> Qarray:
    """Quantum Master Equation solver.

    Args:
        H: time dependent Hamiltonian function or time-independent Qarray.
        rho0: initial state, must be a density matrix. For statevector evolution, please use sesolve.
        tlist: time list
        c_ops: qarray list of collapse operators
        solver_options: SolverOptions with solver options

    Returns:
        list of states
    """
    
    c_ops = c_ops if c_ops is not None else Qarray.from_list([])

    # if isinstance(H, Qarray):
        

    if len(c_ops) == 0 and rho0.qtype != Qtypes.oper:
        logging.warning(
            "Consider using `jqt.sesolve()` instead, as `c_ops` is an empty list and the initial state is not a density matrix."
        )

    ρ0 = rho0.to_dm()
    dims = ρ0.dims
    ρ0 = ρ0.data

    c_ops = c_ops.data

    if isinstance(H, Qarray):
        Ht_data = lambda t: H.data
    else:
        Ht_data = lambda t: H(t).data if H is not None else None
    
    ys = _mesolve_data(Ht_data, ρ0, tlist, c_ops, solver_options=solver_options)

    return jnp2jqt(ys, dims=dims)


def _mesolve_data(
    H: Callable[[float], Array],
    rho0: Array,
    tlist: Array,
    c_ops: Optional[Qarray] = None,
    solver_options: Optional[SolverOptions] = None
) -> Array:
    """Quantum Master Equation solver.

    Args:
        H: time dependent Hamiltonian function or time-independent Array.
        rho0: initial state, must be a density matrix. For statevector evolution, please use sesolve.
        tlist: time list
        c_ops: qarray list of collapse operators
        solver_options: SolverOptions with solver options

    Returns:
        list of states
    """
    
    c_ops = c_ops if c_ops is not None else jnp.array([])

    # check is in mesolve
    # if len(c_ops) == 0 and not is_dm_data(rho0):
    #     logging.warning(
    #         "Consider using `jqt.sesolve()` instead, as `c_ops` is an empty list and the initial state is not a density matrix."
    #     )

    ρ0 = rho0 + 0.0j

    if len(c_ops) == 0:
        test_data = H(0.0) @ ρ0
    else:
        test_data = c_ops[0] @ H(0.0) @ ρ0
    
    ρ0 = jnp.resize(ρ0, test_data.shape)  # ensure correct shape

    if len(c_ops) != 0:
        c_ops_bdims = c_ops.shape[:-2]
        c_ops = c_ops.reshape(*c_ops_bdims, c_ops.shape[-2], c_ops.shape[-1])

    def f(
        t: float,
        rho: Array,
        c_ops_val: Array,
    ):

        H_val = H(t)  # type: ignore
        H_val = H_val + 0.0j

        rho_dot = -1j * (H_val @ rho - rho @ H_val)

        if len(c_ops_val) == 0:
            return rho_dot

        c_ops_val_dag = dag_data(c_ops_val)

        rho_dot_delta = (0.5 * (
            2 * c_ops_val @ rho @ c_ops_val_dag
            - rho @ c_ops_val_dag @ c_ops_val
            - c_ops_val_dag @ c_ops_val @ rho
        ))

        rho_dot_delta = jnp.sum(rho_dot_delta, axis=0)

        rho_dot += rho_dot_delta
        
        return rho_dot

    sol = solve(f, ρ0, tlist, c_ops, solver_options=solver_options)

    return sol.ys

def sesolve(
    H: Union[Qarray, Callable[[float], Qarray]],
    rho0: Qarray,
    tlist: Array,
    solver_options: Optional[SolverOptions] = None
) -> Qarray:
    """Schrödinger Equation solver.

    Args:
        H: time dependent Hamiltonian function or time-independent Qarray.
        rho0: initial state, must be a density matrix. For statevector evolution, please use sesolve.
        tlist: time list
        solver_options: SolverOptions with solver options

    Returns:
        list of states
    """

    ψ = rho0

    if ψ.qtype == Qtypes.oper:
        raise ValueError(
            "Please use `jqt.mesolve` for initial state inputs in density matrix form."
        )
    
    ψ = ψ.to_ket()
    dims = ψ.dims
    ψ = ψ.data 

    if isinstance(H, Qarray):
        Ht_data = lambda t: H.data
    else:
        Ht_data = lambda t: H(t).data if H is not None else None
    
    ys = _sesolve_data(Ht_data, ψ, tlist, solver_options=solver_options)

    return jnp2jqt(ys, dims=dims)

def _sesolve_data(
    H: Callable[[float], Array],
    rho0: Array,
    tlist: Array,
    solver_options: Optional[SolverOptions] = None
):
    """Schrödinger Equation solver.

    Args:
        H: time dependent Hamiltonian function or time-independent Array.
        rho0: initial state, must be a density matrix. For statevector evolution, please use sesolve.
        tlist: time list
        solver_options: SolverOptions with solver options

    Returns:
        list of states
    """
    
    ψ = rho0
    ψ = ψ + 0.0j

    def f(
        t: float,
        ψₜ: Array,
        _
    ):

        H_val = H(t)  # type: ignore
        H_val = H_val + 0.0j
        
        ψₜ_dot = -1j * (H_val @ ψₜ)

        return ψₜ_dot

    
    ψ_test = f(0, ψ, None)
    ψ = jnp.resize(ψ, ψ_test.shape)  # ensure correct shape

    sol = solve(f, ψ, tlist, None, solver_options=solver_options)
    return sol.ys

# ----


# propagators 
# ----


# def propagator(
#     H: Union[Qarray, Callable[[float], Qarray]],
#     t: Union[float, Array],
#     solver_options=None
# ):
#     """ Generate the propagator for a time dependent Hamiltonian. 

#     Args:
#         H (Qarray or callable): 
#             A Qarray static Hamiltonian OR
#             a function that takes a time argument and returns a Hamiltonian. 
#         ts (float or Array): 
#             A single time point or
#             an Array of time points.
        
#     Returns:
#         Qarray or List[Qarray]: 
#             The propagator for the Hamiltonian at time t.
#             OR a list of propagators for the Hamiltonian at each time in t.

#     """

#     t_is_scalar = robust_isscalar(t)

#     if isinstance(H, Qarray):
#         dims = H.dims 
#         if t_is_scalar:
#             if t == 0:
#                 return identity_like(H)

#             return jnp2jqt(propagator_0_data(H.data,t), dims=dims)
#         else:
#             f = lambda t: propagator_0_data(H.data,t)
#             return jnp2jqt(vmap(f)(t), dims)
#     else:
#         dims = H(0.0).dims
#         H_data = lambda t: H(t).data
#         if t_is_scalar:
#             if t == 0:
#                 return identity_like(H(0.0))

#             ts = jnp.linspace(0,t,2)
#             return jnp2jqt(
#                 propagator_t_data(H_data, ts, solver_options=solver_options)[1],
#                 dims=dims
#             )
#         else:
#             ts = t 
#             U_props = propagator_t_data(H_data, ts, solver_options=solver_options)
#             return jnp2jqt(U_props, dims)

# def propagator_0_data(
#     H0: Array,
#     t: float
# ):
#     """ Generate the propagator for a time independent Hamiltonian. 

#     Args:
#         H0 (Qarray): The Hamiltonian.

#     Returns:
#         Qarray: The propagator for the time independent Hamiltonian.
#     """
#     return jsp.linalg.expm(-1j * H0 * t)

# def propagator_t_data(
#     Ht: Callable[[float], Array],
#     ts: Array, 
#     solver_options=None
# ):
    """ Generate the propagator for a time dependent Hamiltonian. 

    Args:
        ts (float): The final time of the propagator. 
            Warning: Do not send in t. In this case, just do exp(-1j*Ht(0.0)).
        Ht (callable): A function that takes a time argument and returns a Hamiltonian. 
        solver_options (dict): Options to pass to the solver.

    Returns:
        Qarray: The propagator for the time dependent Hamiltonian for the time range [0, t_final].
    """
    # N = Ht(0).shape[0]
    # basis_states = jnp.eye(N)

    # def propogate_state(initial_state):
    #     return sesolve_data(initial_state, ts, Ht=Ht, solver_options=solver_options)
        
    # U_prop = vmap(propogate_state)(basis_states)
    # U_prop = U_prop.transpose(1,0,2) # move time axis to the front
    # return U_prop


"""
Cat Code Qubit
"""

from typing import Tuple


from jaxquantum.utils.utils import comb
from jaxquantum.codes.base import BosonicQubit
import jaxquantum as jqt

from jax import vmap
from jax import config
import jax.numpy as jnp

config.update("jax_enable_x64", True)


class BinomialQubit(BosonicQubit):
    """
    Cat Qubit Class.
    """
    name = "binomial"

    def _params_validation(self):
        super()._params_validation()

        # notation https://arxiv.org/pdf/2010.08699.pdf
        if "L" not in self.params:
            self.params["L"] = 1
        if "G" not in self.params:
            self.params["G"] = 0
        if "D" not in self.params:
            self.params["D"] = 0

    def _get_basis_z(self) -> Tuple[jqt.Qarray, jqt.Qarray]:
        """
        Construct basis states |+-x>, |+-y>, |+-z>
        """
        N = self.params["N"]

        L = self.params["L"]
        G = self.params["G"]
        D = self.params["D"]

        S = L + G

        M = jnp.max(jnp.array([L, G, 2 * D]))
        
        def plus_z_gen(p):
            C = comb(M + 1, p)
            return jnp.sqrt(C) * jqt.basis(N, p * (S + 1)).data

        plus_z = jnp.sum(vmap(plus_z_gen)(jnp.arange(0, M + 2, 2)), axis=0)
        plus_z = jqt.unit(jqt.Qarray.create(plus_z))

        def minus_z_gen(p):
            C = comb(M + 1, p)
            return jnp.sqrt(C) * jqt.basis(N, p * (S + 1)).data

        minus_z = jnp.sum(vmap(minus_z_gen)(jnp.arange(1, M + 2, 2)), axis=0)
        minus_z = jqt.unit(jqt.Qarray.create(minus_z))

        return plus_z, minus_z




""" Base device."""

from abc import abstractmethod, ABC
from enum import Enum
from typing import Dict, Any, List

from flax import struct
from jax import config, Array
import jax.numpy as jnp
import matplotlib.pyplot as plt

from jaxquantum.devices.common.utils import harm_osc_wavefunction
from jaxquantum.core.qarray import Qarray
from jaxquantum.core.dims import Qtypes

config.update("jax_enable_x64", True)


class BasisTypes(str, Enum):
    fock = "fock"
    charge = "charge"
    single_charge = "single_charge"

    @classmethod
    def from_str(cls, string: str):
        return cls(string)

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return self.value != other.value

    def __hash__(self):
        return hash(self.value)

class HamiltonianTypes(str, Enum):
    linear = "linear"
    truncated = "truncated"
    full = "full"

    @classmethod
    def from_str(cls, string: str):
        return cls(string)

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return self.value != other.value

    def __hash__(self):
        return hash(self.value)

@struct.dataclass
class Device(ABC):
    DEFAULT_BASIS = BasisTypes.fock
    DEFAULT_HAMILTONIAN = HamiltonianTypes.full

    N: int = struct.field(pytree_node=False)
    N_pre_diag: int = struct.field(pytree_node=False)
    params: Dict[str, Any]
    _label: int = struct.field(pytree_node=False)
    _basis: BasisTypes = struct.field(pytree_node=False)
    _hamiltonian: HamiltonianTypes = struct.field(pytree_node=False)

    @classmethod
    def param_validation(cls, N, N_pre_diag, params, hamiltonian, basis):
        """ This can be overridden by subclasses."""
        pass

    @classmethod
    def create(cls, N, params, label=0, N_pre_diag=None,use_linear=False, hamiltonian: HamiltonianTypes = None, basis: BasisTypes = None):
        """Create a device.

        Args:
            N (int): dimension of Hilbert space.
            params (dict): parameters of the device.
            label (int, optional): label for the device. Defaults to 0. This is useful when you have multiple of the same device type in the same system.
            N_pre_diag (int, optional): dimension of Hilbert space before diagonalization. Defaults to None, in which case it is set to N. This must be greater than or rqual to N.
            use_linear (bool): whether to use the linearized device. Defaults to False. This will override the hamiltonian keyword argument. This is a bit redundant with hamiltonian, but it is kept for backwards compatibility.
            hamiltonian (HamiltonianTypes, optional): type of Hamiltonian. Defaults to None, in which case the full hamiltonian is used.
            basis (BasisTypes, optional): type of basis. Defaults to None, in which case the fock basis is used.
        """
        
        if N_pre_diag is None:
            N_pre_diag = N

        assert N_pre_diag >= N, "N_pre_diag must be greater than or equal to N."

        _basis = basis if basis is not None else cls.DEFAULT_BASIS
        _hamiltonian = hamiltonian if hamiltonian is not None else cls.DEFAULT_HAMILTONIAN
        
        if use_linear:
            _hamiltonian = HamiltonianTypes.linear
        
        cls.param_validation(N, N_pre_diag, params, _hamiltonian, _basis)

        return cls(N, N_pre_diag, params, label, _basis, _hamiltonian)
    
    @property
    def basis(self):
        return self._basis
    
    @property
    def hamiltonian(self):
        return self._hamiltonian

    @property
    def label(self):
        return self.__class__.__name__ + str(self._label)

    @property
    def linear_ops(self):
        return self.common_ops()
    
    @property
    def original_ops(self):
        return self.common_ops()

    @property
    def ops(self):
        return self.full_ops()

    @abstractmethod
    def common_ops(self) -> Dict[str, Qarray]:
        """Set up common ops in the specified basis."""

    @abstractmethod
    def get_linear_ω(self):
        """Get frequency of linear terms."""

    @abstractmethod
    def get_H_linear(self):
        """Return linear terms in H."""

    @abstractmethod
    def get_H_full(self):
        """Return full H."""

    def get_H(self):
        """
        Return diagonalized H. Explicitly keep only diagonal elements of matrix.
        """
        return self.get_op_in_H_eigenbasis(self._get_H_in_original_basis()).keep_only_diag_elements()

    def _get_H_in_original_basis(self):
        """ This returns the Hamiltonian in the original specified basis. This can be overridden by subclasses."""

        if self.hamiltonian == HamiltonianTypes.linear:
            return self.get_H_linear()
        elif self.hamiltonian == HamiltonianTypes.full:
            return self.get_H_full()
        
    def _calculate_eig_systems(self):
        evs, evecs = jnp.linalg.eigh(self._get_H_in_original_basis().data)  # Hermitian
        idxs_sorted = jnp.argsort(evs)
        return evs[idxs_sorted], evecs[:, idxs_sorted]

    @property
    def eig_systems(self):
        eig_systems = {}
        eig_systems["vals"], eig_systems["vecs"] = self._calculate_eig_systems()

        eig_systems["vecs"] = eig_systems["vecs"]
        eig_systems["vals"] = eig_systems["vals"]
        return eig_systems

    def get_op_in_H_eigenbasis(self, op: Qarray):
        evecs = self.eig_systems["vecs"][:, : self.N]
        dims = [[self.N], [self.N]]
        return get_op_in_new_basis(op, evecs, dims)
    
    def get_op_data_in_H_eigenbasis(self, op: Array):
        evecs = self.eig_systems["vecs"][:, : self.N]
        return get_op_data_in_new_basis(op, evecs)

    def get_vec_in_H_eigenbasis(self, vec: Qarray):
        evecs = self.eig_systems["vecs"][:, : self.N]
        if vec.qtype == Qtypes.ket:
            dims = [[self.N],[1]]
        else:
            dims = [[1], [self.N]]
        return get_vec_in_new_basis(vec, evecs, dims)
    
    def get_vec_data_in_H_eigenbasis(self, vec: Array):
        evecs = self.eig_systems["vecs"][:, : self.N]
        return get_vec_data_in_new_basis(vec, evecs)

    def full_ops(self):
        # TODO: use JAX vmap here

        linear_ops = self.linear_ops
        ops = {}
        for name, op in linear_ops.items():
            ops[name] = self.get_op_in_H_eigenbasis(op)

        return ops


def get_op_in_new_basis(op: Qarray, evecs: Array, dims: List[List[int]]) -> Qarray:
    data = get_op_data_in_new_basis(op.data, evecs)
    return Qarray.create(data, dims=dims)

def get_op_data_in_new_basis(op_data: Array, evecs: Array) -> Array:
    return jnp.dot(jnp.conjugate(evecs.transpose()), jnp.dot(op_data, evecs))

def get_vec_in_new_basis(vec: Qarray, evecs: Array, dims: List[List[int]]) -> Qarray:
    return Qarray.create(get_vec_data_in_new_basis(vec.data, evecs), dims=dims)

def get_vec_data_in_new_basis(vec_data: Array, evecs: Array) -> Array:
    return jnp.dot(jnp.conjugate(evecs.transpose()), vec_data)
